import asyncio
import threading
import traceback
import uuid
import warnings

from typing import Any, Callable

import structlog

from ...client.abstractions import (
    SSE,
    ConsumedThingAction,
    ConsumedThingEvent,
    ConsumedThingProperty,
    raise_local_exception,
)
from ...constants import Operations
from ...core import Action, Thing  # noqa: F401
from ...core.zmq.brokers import (
    AsyncEventConsumer,
    AsyncZMQClient,
    BreakLoop,
    EventConsumer,
    SyncZMQClient,
)
from ...core.zmq.message import (
    EMPTY_BYTE,
    ERROR,
    INVALID_MESSAGE,
    TIMEOUT,
    ResponseMessage,
)
from ...serializers.payloads import SerializableData
from ...td import ActionAffordance, EventAffordance, PropertyAffordance
from ...td.forms import Form
from ..exceptions import ReplyNotArrivedError


__error_message_types__ = [TIMEOUT, ERROR, INVALID_MESSAGE]


class ZMQConsumedAffordanceMixin:
    # A utility mixin class for ZMQ based affordances
    # Dont add doc otherwise __doc__ in slots will conflict with class variable

    __slots__ = [
        "resource",
        "logger",
        "schema_validator",
        "owner_inst",
        "__name__",
        "__qualname__",
        "__doc__",
        "_sync_zmq_client",
        "_async_zmq_client",
        "_invokation_timeout",
        "_execution_timeout",
        "_thing_execution_context",
        "_last_zmq_response",
    ]  # __slots__ dont support multiple inheritance

    def __init__(
        self,
        sync_client: SyncZMQClient,
        async_client: AsyncZMQClient | None = None,
        **kwargs,
    ) -> None:
        """
        Parameters
        ----------
        sync_client: SyncZMQClient
            synchronous ZMQ client
        async_client: AsyncZMQClient
            asynchronous ZMQ client for async calls
        kwargs:
            additional keyword arguments:

            - `invokation_timeout`: float, default 5.0
                timeout for invokation of action or property read/write
            - `execution_timeout`: float, default 5.0
                timeout for execution of action or property read/write
        """
        self._sync_zmq_client = sync_client
        self._async_zmq_client = async_client
        self._invokation_timeout = kwargs.get("invokation_timeout", 5.0)
        self._execution_timeout = kwargs.get("execution_timeout", 5.0)
        self._thing_execution_context = dict(fetch_execution_logs=False)
        self._last_zmq_response = None  # type: ResponseMessage | None

    def get_last_return_value(self, response: ResponseMessage, raise_exception: bool = False) -> Any:
        """
        cached return value of the last operation performed.

        Parameters
        ----------
        response: ResponseMessage
            last response message received from the server
        raise_exception: bool
            whether to raise exception if the last response was an error message
        """
        if response is None:
            raise RuntimeError("No last response available. Did you make an operation?")
        payload = response.payload.deserialize()
        preserialized_payload = response.preserialized_payload.value
        if response.type in __error_message_types__ and raise_exception:
            raise_local_exception(payload)
        if preserialized_payload != EMPTY_BYTE:
            if payload is None:
                return preserialized_payload
            return payload, preserialized_payload
        return payload

    @property
    def last_zmq_response(self) -> ResponseMessage:
        """cache of last message received for this property"""
        return self._last_zmq_response

    def read_reply(self, message_id: str, timeout: int = None) -> Any:
        if self.owner_inst._noblock_messages.get(message_id) != self:
            raise RuntimeError(f"Message ID {message_id} does not belong to this property.")
        response = self._sync_zmq_client.recv_response(message_id=message_id)
        if not response:
            raise ReplyNotArrivedError(f"could not fetch reply within timeout for message id '{message_id}'")
        self._last_zmq_response = response
        return ZMQConsumedAffordanceMixin.get_last_return_value(self, response, True)


class ZMQAction(ZMQConsumedAffordanceMixin, ConsumedThingAction):
    # ZMQ method call abstraction
    # Dont add doc otherwise __doc__ in slots will conflict with class variable

    def __init__(
        self,
        resource: ActionAffordance,
        sync_client: SyncZMQClient,
        async_client: AsyncZMQClient,
        owner_inst: Any,
        logger: structlog.stdlib.BoundLogger,
        **kwargs,
    ) -> None:
        """
        Parameters
        ----------
        resource: ActionAffordance
            dataclass object representing the action
        sync_client: SyncZMQClient
            synchronous ZMQ client
        async_zmq_client: AsyncZMQClient
            asynchronous ZMQ client for async calls
        owner_inst: Any
            the parent object that owns this action
        logger: structlog.stdlib.BoundLogger
            logger instance
        """
        ConsumedThingAction.__init__(self, resource=resource, owner_inst=owner_inst, logger=logger)
        ZMQConsumedAffordanceMixin.__init__(self, sync_client=sync_client, async_client=async_client, **kwargs)
        self.resource = resource

    last_return_value = property(
        fget=lambda self: ZMQConsumedAffordanceMixin.get_last_return_value(self, self._last_zmq_response, True),
        doc="cached return value of the last call to the method",
    )

    def __call__(self, *args, **kwargs) -> Any:
        if len(args) > 0:
            kwargs["__args__"] = args
        elif self.schema_validator:
            self.schema_validator.validate(kwargs)
        form = self.resource.retrieve_form(Operations.invokeaction, Form())
        # works over ThingModel, there can be a default empty form
        response = self._sync_zmq_client.execute(
            thing_id=self.resource.thing_id,
            objekt=self.resource.name,
            operation=Operations.invokeaction,
            payload=SerializableData(value=kwargs, content_type=form.contentType or "application/json"),
            server_execution_context=dict(
                invokation_timeout=self._invokation_timeout,
                execution_timeout=self._execution_timeout,
            ),
            thing_execution_context=self._thing_execution_context,
        )
        self._last_zmq_response = response
        return ZMQConsumedAffordanceMixin.get_last_return_value(self, response, True)

    async def async_call(self, *args, **kwargs) -> Any:
        if not self._async_zmq_client:
            raise RuntimeError("async calls not possible as async_mixin was not set True at __init__()")
        if len(args) > 0:
            kwargs["__args__"] = args
        elif self.schema_validator:
            self.schema_validator.validate(kwargs)
        response = await self._async_zmq_client.async_execute(
            thing_id=self.resource.thing_id,
            objekt=self.resource.name,
            operation=Operations.invokeaction,
            payload=SerializableData(
                value=kwargs,
                content_type=self.resource.retrieve_form(Operations.invokeaction, Form()).contentType
                or "application/json",
            ),
            server_execution_context=dict(
                invokation_timeout=self._invokation_timeout,
                execution_timeout=self._execution_timeout,
            ),
            thing_execution_context=self._thing_execution_context,
        )
        self._last_zmq_response = response
        return ZMQConsumedAffordanceMixin.get_last_return_value(self, response, True)

    def oneway(self, *args, **kwargs) -> None:
        if len(args) > 0:
            kwargs["__args__"] = args
        elif self.schema_validator:
            self.schema_validator.validate(kwargs)
        self._sync_zmq_client.send_request(
            thing_id=self.resource.thing_id,
            objekt=self.resource.name,
            operation=Operations.invokeaction,
            payload=SerializableData(
                value=kwargs,
                content_type=self.resource.retrieve_form(Operations.invokeaction, Form()).contentType
                or "application/json",
            ),
            server_execution_context=dict(
                invokation_timeout=self._invokation_timeout,
                execution_timeout=self._execution_timeout,
                oneway=True,
            ),
            thing_execution_context=self._thing_execution_context,
        )

    def noblock(self, *args, **kwargs) -> str:
        if len(args) > 0:
            kwargs["__args__"] = args
        elif self.schema_validator:
            self.schema_validator.validate(kwargs)
        msg_id = self._sync_zmq_client.send_request(
            thing_id=self.resource.thing_id,
            objekt=self.resource.name,
            operation=Operations.invokeaction,
            payload=SerializableData(
                value=kwargs,
                content_type=self.resource.retrieve_form(Operations.invokeaction, Form()).contentType
                or "application/json",
            ),
            server_execution_context=dict(
                invokation_timeout=self._invokation_timeout,
                execution_timeout=self._execution_timeout,
            ),
            thing_execution_context=self._thing_execution_context,
        )
        self.owner_inst._noblock_messages[msg_id] = self
        return msg_id


class ZMQProperty(ZMQConsumedAffordanceMixin, ConsumedThingProperty):
    # property get set abstraction
    # Dont add doc otherwise __doc__ in slots will conflict with class variable

    def __init__(
        self,
        resource: PropertyAffordance,
        sync_client: SyncZMQClient,
        async_client: AsyncZMQClient,
        owner_inst: Any,
        logger: structlog.stdlib.BoundLogger,
        **kwargs,
    ) -> None:
        """
        Parameters
        ----------
        resource: PropertyAffordance
            dataclass object representing the property
        sync_client: SyncZMQClient
            synchronous ZMQ client
        async_client: AsyncZMQClient
            asynchronous ZMQ client for async calls
        owner_inst: Any
            the parent object that owns this property
        logger: structlog.stdlib.BoundLogger
            logger instance for logging
        """
        ConsumedThingProperty.__init__(self, resource=resource, owner_inst=owner_inst, logger=logger)
        ZMQConsumedAffordanceMixin.__init__(self, sync_client=sync_client, async_client=async_client, **kwargs)
        self.resource = resource

    last_read_value = property(
        fget=lambda self: ZMQConsumedAffordanceMixin.get_last_return_value(self, self._last_zmq_response, True),
        doc="cached return value of the last call to the method",
    )

    def set(self, value: Any) -> None:
        response = self._sync_zmq_client.execute(
            thing_id=self.resource.thing_id,
            objekt=self.resource.name,
            operation=Operations.writeproperty,
            payload=SerializableData(
                value=value,
                content_type=self.resource.retrieve_form(Operations.writeproperty, Form()).contentType
                or "application/json",
            ),
            server_execution_context=dict(
                invokation_timeout=self._invokation_timeout,
                execution_timeout=self._execution_timeout,
            ),
            thing_execution_context=self._thing_execution_context,
        )
        self._last_zmq_response = response
        ZMQConsumedAffordanceMixin.get_last_return_value(self, response, True)

    def get(self) -> Any:
        response = self._sync_zmq_client.execute(
            thing_id=self.resource.thing_id,
            objekt=self.resource.name,
            operation=Operations.readproperty,
            server_execution_context=dict(
                invocation_timeout=self._invokation_timeout,
                execution_timeout=self._execution_timeout,
            ),
            thing_execution_context=self._thing_execution_context,
        )
        self._last_zmq_response = response
        return ZMQConsumedAffordanceMixin.get_last_return_value(self, response, True)

    async def async_set(self, value: Any) -> None:
        if not self._async_zmq_client:
            raise RuntimeError("async calls not possible as async_mixin was not set at __init__()")
        response = await self._async_zmq_client.async_execute(
            thing_id=self.resource.thing_id,
            objekt=self.resource.name,
            operation=Operations.writeproperty,
            payload=SerializableData(
                value=value,
                content_type=self.resource.retrieve_form(Operations.writeproperty, Form()).contentType
                or "application/json",
            ),
            server_execution_context=dict(
                invokation_timeout=self._invokation_timeout,
                execution_timeout=self._execution_timeout,
            ),
            thing_execution_context=self._thing_execution_context,
        )
        self._last_zmq_response = response
        ZMQConsumedAffordanceMixin.get_last_return_value(self, response, True)

    async def async_get(self) -> Any:
        if not self._async_zmq_client:
            raise RuntimeError("async calls not possible as async_mixin was not set at __init__()")
        response = await self._async_zmq_client.async_execute(
            thing_id=self.resource.thing_id,
            objekt=self.resource.name,
            operation=Operations.readproperty,
            server_execution_context=dict(
                invokation_timeout=self._invokation_timeout,
                execution_timeout=self._execution_timeout,
            ),
            thing_execution_context=self._thing_execution_context,
        )
        self._last_zmq_response = response
        return ZMQConsumedAffordanceMixin.get_last_return_value(self, response, True)

    def oneway_set(self, value: Any) -> None:
        self._sync_zmq_client.send_request(
            thing_id=self.resource.thing_id,
            objekt=self.resource.name,
            operation=Operations.writeproperty,
            payload=SerializableData(
                value=value,
                content_type=self.resource.retrieve_form(Operations.writeproperty, Form()).contentType
                or "application/json",
            ),
            server_execution_context=dict(
                invokation_timeout=self._invokation_timeout,
                execution_timeout=self._execution_timeout,
                oneway=True,
            ),
        )

    def noblock_get(self) -> str:
        msg_id = self._sync_zmq_client.send_request(
            thing_id=self.resource.thing_id,
            objekt=self.resource.name,
            operation=Operations.readproperty,
            server_execution_context=dict(
                invokation_timeout=self._invokation_timeout,
                execution_timeout=self._execution_timeout,
            ),
            thing_execution_context=self._thing_execution_context,
        )
        self.owner_inst._noblock_messages[msg_id] = self
        return msg_id

    def noblock_set(self, value: Any) -> None:
        msg_id = self._sync_zmq_client.send_request(
            thing_id=self.resource.thing_id,
            objekt=self.resource.name,
            operation=Operations.writeproperty,
            payload=SerializableData(
                value=value,
                content_type=self.resource.retrieve_form(Operations.writeproperty, Form()).contentType
                or "application/json",
            ),
            server_execution_context=dict(
                invokation_timeout=self._invokation_timeout,
                execution_timeout=self._execution_timeout,
            ),
            thing_execution_context=self._thing_execution_context,
        )
        self.owner_inst._noblock_messages[msg_id] = self
        return msg_id


class ZMQEvent(ConsumedThingEvent, ZMQConsumedAffordanceMixin):
    # Dont add class doc otherwise __doc__ in slots will conflict with class variable

    __slots__ = [
        "_subscribed",
    ]

    def __init__(
        self,
        resource: EventAffordance,
        logger: structlog.stdlib.BoundLogger,
        owner_inst: Any,
        **kwargs,
    ) -> None:
        ConsumedThingEvent.__init__(self, resource=resource, logger=logger, owner_inst=owner_inst)
        ZMQConsumedAffordanceMixin.__init__(self, sync_client=None, async_client=None, **kwargs)

    def listen(self, form: Form, callbacks: list[Callable], concurrent: bool, deserialize: bool) -> None:
        sync_event_client = EventConsumer(
            id=f"{self.resource.thing_id}|{self.resource.name}|sync|{uuid.uuid4().hex[:8]}",
            event_unique_identifier=f"{self.resource.thing_id}/{self.resource.name}",
            access_point=form.href,
            logger=self.logger,
        )
        sync_event_client.subscribe()
        task_id = threading.get_ident()
        self._subscribed[task_id] = (True, sync_event_client)
        while True:
            try:
                if not self._subscribed.get(task_id, (False, None))[0]:
                    break
                event_message = sync_event_client.receive(raise_interrupt_as_exception=True)
                if not event_message:
                    continue
                self._last_zmq_response = event_message
                event_data = SSE()
                event_data.id = event_message.id
                event_data.data = self.get_last_return_value(event_message, raise_exception=True)
                self.schedule_callbacks(callbacks, event_data, concurrent)
            except BreakLoop:
                break
            except Exception as ex:
                # traceback.print_exc()
                # TODO: some minor bug here within the zmq receive loop when the loop is interrupted
                # uncomment the above line to see the traceback
                warnings.warn(
                    f"Uncaught exception from {self.resource.name} event - {str(ex)}\n{traceback.print_exc()}",
                    category=RuntimeWarning,
                )

    async def async_listen(self, form: Form, callbacks: list[Callable], concurrent: bool, deserialize: bool) -> None:
        async_event_client = AsyncEventConsumer(
            id=f"{self.resource.thing_id}|{self.resource.name}|async|{uuid.uuid4().hex[:8]}",
            event_unique_identifier=f"{self.resource.thing_id}/{self.resource.name}",
            access_point=form.href,
            logger=self.logger,
        )
        async_event_client.subscribe()
        task_id = asyncio.current_task().get_name()
        self._subscribed[task_id] = (True, async_event_client)
        while True:
            try:
                if not self._subscribed.get(task_id, (False, None))[0]:
                    break
                event_message = await async_event_client.receive(raise_interrupt_as_exception=True)
                if not event_message:
                    continue
                self._last_zmq_response = event_message
                event_data = SSE()
                event_data.id = event_message.id
                event_data.data = self.get_last_return_value(event_message, raise_exception=True)
                await self.async_schedule_callbacks(callbacks, event_data, concurrent)
            except BreakLoop:
                break
            except Exception as ex:
                # traceback.print_exc()
                # if "There is no current event loop in thread" and not self._subscribed:
                #     # TODO: some minor bug here within the umq receive loop when the loop is interrupted
                #     # uncomment the above line to see the traceback
                #    pass
                # else:
                warnings.warn(
                    f"Uncaught exception from {self.resource.name} event - {str(ex)}\n{traceback.print_exc()}",
                    category=RuntimeWarning,
                )

    def unsubscribe(self) -> None:
        for task_id, (subscribed, client) in self._subscribed.items():
            if client:
                client.stop_polling()
        return super().unsubscribe()


class WriteMultipleProperties(ZMQAction):
    """
    Read and write multiple properties at once
    """

    def __init__(
        self,
        sync_client: SyncZMQClient,
        async_client: AsyncZMQClient | None = None,
        owner_inst: Any = None,
        **kwargs,
    ) -> None:
        action = Thing._set_properties  # type: Action
        resource = action.to_affordance(Thing)
        resource._thing_id = owner_inst.thing_id
        super().__init__(
            resource=resource,
            sync_client=sync_client,
            async_client=async_client,
            owner_inst=owner_inst,
            **kwargs,
        )


class ReadMultipleProperties(ZMQAction):
    """
    Read multiple properties at once
    """

    def __init__(
        self,
        sync_client: SyncZMQClient,
        async_client: AsyncZMQClient | None = None,
        owner_inst: Any = None,
        **kwargs,
    ) -> None:
        action = Thing._get_properties  # type: Action
        resource = action.to_affordance(Thing)
        resource._thing_id = owner_inst.thing_id
        super().__init__(
            resource=resource,
            sync_client=sync_client,
            async_client=async_client,
            owner_inst=owner_inst,
            **kwargs,
        )


__all__ = [
    ZMQAction.__name__,
    ZMQProperty.__name__,
    ZMQEvent.__name__,
]



from ..config import global_config  # noqa: F401
from .factory import ClientFactory as ClientFactory
from .proxy import ObjectProxy as ObjectProxy


from .security import (  # noqa: F401, isort: skip
    APIKeySecurity,
    BasicSecurity,
    OAuthDirectAccessGrant,
)



class BreakInnerLoop(Exception):
    """raise to break an inner loop"""

    pass


class BreakAllLoops(Exception):
    """raise to exit all loops"""

    pass


class BreakLoop(Exception):
    """raise and catch to exit a loop from within another function or method"""

    pass


class BreakFlow(Exception):
    """raised to break the flow of the program"""

    pass


# TODO - remove unused and reduce number of definitions


class StateMachineError(Exception):
    """raise to show errors while calling actions or writing properties in wrong state"""

    pass


class DatabaseError(Exception):
    """raise to show database related errors"""


__all__ = ["BreakInnerLoop", "BreakAllLoops", "BreakLoop", "BreakFlow", "StateMachineError", "DatabaseError"]



from typing import Any, Optional

import msgspec
import structlog

from msgspec import DecodeError as MsgspecJSONDecodeError
from tornado.iostream import StreamClosedError
from tornado.web import RequestHandler

from ...config import global_config
from ...constants import Operations
from ...core.zmq.brokers import EventConsumer
from ...core.zmq.message import (
    SerializableNone,
    ServerExecutionContext,
    ThingExecutionContext,
    default_server_execution_context,
    default_thing_execution_context,
)
from ...serializers import Serializers
from ...serializers.payloads import PreserializedData, SerializableData
from ...td import (
    ActionAffordance,
    EventAffordance,
    InteractionAffordance,
    PropertyAffordance,
)
from ...utils import format_exception_as_json, get_current_async_loop
from ..repository import BrokerThing  # noqa: F401
from ..security import (
    APIKeySecurity,
    Argon2BasicSecurity,
    BcryptBasicSecurity,
    OIDCSecurity,
)


class LocalExecutionContext(msgspec.Struct):
    noblock: Optional[bool] = None
    messageID: Optional[str] = None


class BaseHandler(RequestHandler):
    """Base request handler for running operations on the `Thing`"""

    # Would be a Controller in layered architecture.

    def initialize(
        self,
        resource: InteractionAffordance | PropertyAffordance | ActionAffordance | EventAffordance,
        config: Any,
        logger: structlog.stdlib.BoundLogger,
        metadata: Any = None,
    ) -> None:
        """
        Parameters
        ----------
        resource: InteractionAffordance | PropertyAffordance | ActionAffordance | EventAffordance
            dataclass representation of `Thing`'s exposed object that can quickly convert to a ZMQ Request object
        owner_inst: HTTPServer
            owning `hololinked.server.HTTPServer` instance
        metadata: HandlerMetadata | None,
            additional metadata about the resource, like allowed HTTP methods
        """
        from .config import HandlerMetadata, RuntimeConfig  # noqa: F401

        self.resource = resource  # type: InteractionAffordance | PropertyAffordance | ActionAffordance | EventAffordance
        self.config = config  # type: RuntimeConfig
        self.logger = logger.bind(
            resource=resource.name,
            what=resource.what,
            thing_id=resource.thing_id,
            path=self.request.path,
            layer="controller",
            impl=self.__class__.__name__,
        )
        self.thing = self.config.thing_repository[self.resource.thing_id]  # type: BrokerThing
        self.allowed_clients = self.config.allowed_clients
        self.security_schemes = self.config.security_schemes
        self.metadata = metadata or HandlerMetadata()  # type: HandlerMetadata
        self.userinfo = None  # type: Optional[dict[str, Any]]

    async def has_access_control(self) -> bool:
        """
        Checks if a client is an allowed client and enforces security schemes.
        Custom web request handlers can use this property to check if a client has access control on the server or `Thing`
        and let this property automatically generate a 401/403.
        """
        if not self.allowed_clients and not self.security_schemes:
            return True
        # First check if the client is allowed to access the server
        origin = self.request.headers.get("Origin")
        if (
            self.allowed_clients
            and origin is not None
            and (origin not in self.allowed_clients and origin + "/" not in self.allowed_clients)
        ):
            self.set_status(401, "Unauthorized")
            return False
        # Then check an authentication scheme either if the client is allowed
        # or if there is no such list of allowed clients
        if not self.security_schemes:
            self.logger.debug("no security schemes defined, allowing access")
            return True
        if await self.is_authenticated():
            self.logger.info("client authenticated successfully")
            if not self.userinfo:
                return True
            if await self.is_authorized():
                return True
            self.set_status(403, "Forbidden")
            self.logger.info("insufficient permissions")
            return False
        self.set_status(401, "Unauthorized")
        self.logger.info("client authentication failed or is not authorized to proceed")
        return False  # keep False always at the end

    async def is_authenticated(self) -> bool:
        """enforces authentication using the defined security schemes, freshly computed everytime"""
        authenticated = False
        # 1. Basic Authentication
        authorization_header = self.request.headers.get("Authorization", None)  # type: str
        if authorization_header and "basic " in authorization_header[:10].lower():  # basic <base64-encoded>
            for security_scheme in self.security_schemes:
                if isinstance(security_scheme, (BcryptBasicSecurity, Argon2BasicSecurity)):
                    try:
                        self.logger.info(
                            "authenticating client",
                            origin=self.request.headers.get("Origin"),
                            security_scheme=security_scheme.__class__.__name__,
                        )
                        if security_scheme.expect_base64:
                            authenticated = security_scheme.validate_base64(authorization_header.split()[1])
                        else:
                            authenticated = security_scheme.validate_input(
                                username=authorization_header.split()[1].split(":", 1)[0],
                                password=authorization_header.split()[1].split(":", 1)[1],
                            )
                    except Exception as ex:
                        self.logger.error(f"error while authenticating client - {str(ex)}")
                    if authenticated:
                        return True
        # 2. API Key Authentication
        apikey = self.request.headers.get("X-API-Key", None)  # type: str
        if apikey:
            for security_scheme in self.security_schemes:
                if isinstance(security_scheme, APIKeySecurity):
                    try:
                        self.logger.info(
                            "authenticating client with API key",
                            origin=self.request.headers.get("Origin"),
                            security_scheme=security_scheme.__class__.__name__,
                        )
                        authenticated = security_scheme.validate_input(apikey)
                    except Exception as ex:
                        self.logger.error(f"error while authenticating client with API key - {str(ex)}")
                    if authenticated:
                        return True
        # 3. JWT from OIDC
        if authorization_header and "bearer " in authorization_header[:10].lower():  # bearer <bla-bla>
            for security_scheme in self.security_schemes:
                if isinstance(security_scheme, OIDCSecurity):
                    try:
                        self.logger.info(
                            "authenticating client with OIDC JWT",
                            origin=self.request.headers.get("Origin"),
                            security_scheme=security_scheme.__class__.__name__,
                        )
                        jwt = authorization_header.split(maxsplit=1)[1]
                        self.userinfo = security_scheme.validate_input(jwt)
                        # local validation of JWT, no await
                        if not self.userinfo:
                            continue
                        authenticated = True
                    except Exception as ex:
                        self.logger.error(f"error while authenticating client with OIDC JWT - {str(ex)}")
                    if authenticated:
                        return True
        return authenticated

    async def is_authorized(self) -> bool:
        """
        enforces authorization using the defined security schemes, freshly computed everytime.
        Do not call this method without doing authentication first and having a userinfo property.
        """
        if not self.userinfo:
            return True
        for security_scheme in self.security_schemes:
            if isinstance(security_scheme, OIDCSecurity):
                return security_scheme.user_has_role(self.userinfo)
        return False

    def set_access_control_allow_headers(self) -> None:
        """
        For credential login, access control allow headers cannot be a wildcard '*'.
        Some requests require exact list of allowed headers for the client to access the response.
        """
        headers = ", ".join(self.request.headers.keys())
        if self.request.headers.get("Access-Control-Request-Headers", None):
            headers += ", " + self.request.headers["Access-Control-Request-Headers"]
        self.set_header("Access-Control-Allow-Headers", headers)

    def set_custom_default_headers(self) -> None:
        """
        sets general default headers, override in child classes to add more headers.

        ```yaml
        Content-Type: application/json
        Access-Control-Allow-Origin: <client>
        ```
        """
        # Access-Control-Allow-Credentials: true # only for cookie auth
        if self.config.cors:
            # For credential login, access control allow origin cannot be '*',
            # See: https://developer.mozilla.org/en-US/docs/Web/HTTP/CORS#examples_of_access_control_scenarios
            self.set_header("Access-Control-Allow-Origin", "*")

    def get_execution_parameters(
        self,
    ) -> tuple[
        ServerExecutionContext,
        ThingExecutionContext,
        LocalExecutionContext,
        SerializableData,
    ]:
        """
        Aggregates all arguments to a standard dataclasses from the query parameters.
        Retrieves execution context (like oneway calls, fetching executing
        logs), timeouts, etc. Non recognized arguments are passed as additional payload to the `Thing`.

        An example would be the following URL:

        ```
        http://localhost:8080/property/temperature?oneway=true&invokationTimeout=5&some_arg=42
        ```

        server execution context would have `oneway` set to true & `invokationTimeout` set to 5 seconds,
        local execution context would be empty as no such arguments were passed,
        and additional payload would have `{"some_arg": 42}` as its value.

        Returns
        -------
        tuple[
            ServerExecutionContext,
            ThingExecutionContext,
            LocalExecutionContext,
            SerializableData,
        ]
            server execution context, thing execution context, local execution context and payload (if any)
        """
        arguments = dict()
        if len(self.request.query_arguments) == 0:
            return (
                default_server_execution_context,
                default_thing_execution_context,
                LocalExecutionContext(),
                SerializableNone,
            )
        for key, value in self.request.query_arguments.items():
            if len(value) == 1:
                try:
                    arguments[key] = Serializers.json.loads(value[0])
                except MsgspecJSONDecodeError:
                    arguments[key] = value[0].decode("utf-8")
            else:
                final_value = []
                for val in value:
                    try:
                        final_value.append(Serializers.json.loads(val))
                    except MsgspecJSONDecodeError:
                        final_value.append(val.decode("utf-8"))
                arguments[key] = final_value
        # if self.resource.request_as_argument:
        #     arguments['request'] = self.request # find some way to pass the request object to the thing
        thing_execution_context = ThingExecutionContext(
            fetchExecutionLogs=bool(arguments.pop("fetchExecutionLogs", False))
        )
        server_execution_context = ServerExecutionContext(
            invokationTimeout=arguments.pop("invokationTimeout", default_server_execution_context.invokationTimeout),
            executionTimeout=arguments.pop("executionTimeout", default_server_execution_context.executionTimeout),
            oneway=arguments.pop("oneway", default_server_execution_context.oneway),
        )
        local_execution_context = LocalExecutionContext(
            noblock=arguments.pop("noblock", None),
            messageID=arguments.pop("messageID", None),
        )
        additional_payload = SerializableNone if not arguments else SerializableData(arguments)  # application/json
        return server_execution_context, thing_execution_context, local_execution_context, additional_payload

    @property
    def message_id(self) -> str:
        """retrieves the message id from the request headers"""
        try:
            return self._message_id
        except AttributeError:
            message_id = self.request.headers.get("X-Message-ID", None)
            if not message_id:
                _, _, local_execution_context, _ = self.get_execution_parameters()
                # TODO avoid calling get_execution_parameters twice in the same request
                message_id = local_execution_context.messageID
            self._message_id = message_id
            return message_id

    def get_request_payload(self) -> tuple[SerializableData, PreserializedData]:
        """retrieves the payload from the request body, does not necessarily deserialize it"""
        payload = SerializableData(value=None)
        preserialized_payload = PreserializedData(value=b"")
        if self.request.body:
            if self.request.headers.get("Content-Type", "application/json") in Serializers.allowed_content_types:
                payload.value = self.request.body
                payload.content_type = self.request.headers.get("Content-Type", "application/json")
            elif global_config.ALLOW_UNKNOWN_SERIALIZATION:
                preserialized_payload.value = self.request.body
                preserialized_payload.content_type = self.request.headers.get("Content-Type", None)
            else:
                raise ValueError("Content-Type not supported")
                # NOTE that was assume that the content type is JSON even if unspecified in the header.
                # This error will be raised only when a specified content type is not supported.
        return payload, preserialized_payload

    async def get(self) -> None:
        """runs property or action if accessible by 'GET' method. Default for property reads"""
        raise NotImplementedError("implement GET request method in child handler class")

    async def post(self) -> None:
        """runs property or action if accessible by 'POST' method. Default for action execution"""
        raise NotImplementedError("implement POST request method in child handler class")

    async def put(self) -> None:
        """runs property or action if accessible by 'PUT' method. Default for property writes"""
        raise NotImplementedError("implement PUT request method in child handler class")

    async def delete(self) -> None:
        """
        runs property or action if accessible by 'DELETE' method. Default for property deletes
        (not a valid operation as per web of things semantics).
        """
        raise NotImplementedError("implement DELETE request method in child handler class")

    def is_method_allowed(self, method: str) -> bool:
        """checks if the method is allowed for the property"""
        raise NotImplementedError("implement is_method_allowed in child handler class")


class RPCHandler(BaseHandler):
    """
    Handler for property read-write and method calls.

    Subclassed from controller for reducing boilerplate code as the service layer is too thin when implemented separately.
    Uses Repository layer directly.
    """

    # Merges both Controller and Service layer in layered architecture. Repository layer is used directly.

    async def is_method_allowed(self, method: str) -> bool:
        """
        Checks if the method is allowed for the property:

        - Access control (authentication & authorization)
        - if the HTTP method is allowed for the resource.
        - if its GET method with message id for no-block response.
        """
        if not await self.has_access_control():
            return False
        if self.message_id is not None and method.upper() == "GET":
            return True
        if method in self.metadata.http_methods:
            return True
        self.set_status(405, "method not allowed")
        return False

    async def options(self) -> None:
        """
        Options for the resource. Main functionality is to inform the client is a specific HTTP method is supported by
        the property or the action (Access-Control-Allow-Methods).
        """
        if await self.has_access_control():
            self.set_status(204)
            self.set_custom_default_headers()
            self.set_access_control_allow_headers()
            self.set_header("Access-Control-Allow-Methods", ", ".join(self.metadata.http_methods))
        self.finish()

    async def handle_through_thing(self, operation: str) -> None:
        """
        handles the `Thing` operations and writes the reply to the HTTP client.

        Parameters
        ----------
        operation: str
            operation to be performed on the Thing, like `readproperty`,
            `writeproperty`, `invokeaction`, `deleteproperty`
        """
        try:
            server_execution_context, thing_execution_context, local_execution_context, additional_payload = (
                self.get_execution_parameters()
            )
            payload, preserialized_payload = self.get_request_payload()
            payload = payload if payload.value else additional_payload
        except Exception as ex:
            self.set_status(400, f"error while decoding request - {str(ex)}")
            self.logger.error(f"error while decoding request - {str(ex)}")
            return
        try:
            if server_execution_context.oneway:
                # if oneway, we do not expect a response, so we just return None
                await self.thing.oneway(
                    objekt=self.resource.name,
                    operation=operation,
                    payload=payload,
                    preserialized_payload=preserialized_payload,
                    server_execution_context=server_execution_context,
                    thing_execution_context=thing_execution_context,
                )
                self.set_status(204, "ok")
            elif local_execution_context.noblock:
                message_id = await self.thing.schedule(
                    objekt=self.resource.name,
                    operation=operation,
                    payload=payload,
                    preserialized_payload=preserialized_payload,
                    server_execution_context=server_execution_context,
                    thing_execution_context=thing_execution_context,
                )
                self.set_status(204, "ok")
                self.set_header("X-Message-ID", message_id)
            else:
                response_message = await self.thing.execute(
                    objekt=self.resource.name,
                    operation=operation,
                    payload=payload,
                    preserialized_payload=preserialized_payload,
                    server_execution_context=server_execution_context,
                    thing_execution_context=thing_execution_context,
                )
                response_payload = self.thing.get_response_payload(response_message)
                self.set_status(200, "ok")
                self.set_header("Content-Type", response_payload.content_type or "application/json")
                if response_payload.value:
                    self.write(response_payload.value)
        except ConnectionAbortedError as ex:
            self.set_status(503, f"lost connection to thing - {str(ex)}")
            # TODO handle reconnection
        except Exception as ex:
            self.logger.error(f"error while scheduling RPC call - {str(ex)}")
            self.set_status(500, f"error while scheduling RPC call - {str(ex)}")
            self.set_header("Content-Type", "application/json")
            response_payload = SerializableData(
                value=Serializers.json.dumps({"exception": format_exception_as_json(ex)}),
                content_type="application/json",
            )
            response_payload.serialize()
            self.write(response_payload.value)

    async def handle_no_block_response(self) -> None:
        """handles the no-block response for the noblock calls"""
        try:
            self.logger.info("waiting for no-block response", message_id=self.message_id)
            response_message = await self.thing.recv_response(
                message_id=self.message_id,
                timeout=default_server_execution_context.invokationTimeout
                + default_server_execution_context.executionTimeout,
            )
            response_payload = self.thing.get_response_payload(response_message)
            self.set_status(200, "ok")
            self.set_header("Content-Type", response_payload.content_type or "application/json")
            if response_payload.value:
                self.write(response_payload.value)
        except KeyError as ex:
            # if the message id is not found, it means that the response was not received in time
            self.logger.error(f"message ID not found for no-block response - {str(ex)}")
            self.set_status(404, "message id not found")
        except TimeoutError as ex:
            self.logger.error(f"timeout while waiting for no-block response - {str(ex)}")
            self.set_status(408, "timeout while waiting for response, ask later")
        except Exception as ex:
            self.logger.error(f"error while receiving no-block response - {str(ex)}")
            self.set_status(500, f"error while receiving no-block response - {str(ex)}")
            self.set_header("Content-Type", "application/json")
            response_payload = SerializableData(
                value=Serializers.json.dumps({"exception": format_exception_as_json(ex)}),
                content_type="application/json",
            )
            response_payload.serialize()
            self.write(response_payload.value)


class PropertyHandler(RPCHandler):
    """handles property requests"""

    async def get(self) -> None:
        if await self.is_method_allowed("GET"):
            self.set_custom_default_headers()
            if self.message_id is not None:
                await self.handle_no_block_response()
            else:
                await self.handle_through_thing(Operations.readproperty)
        self.finish()

    async def post(self) -> None:
        if await self.is_method_allowed("POST"):
            self.set_custom_default_headers()
            await self.handle_through_thing(Operations.writeproperty)
        self.finish()

    async def put(self) -> None:
        if await self.is_method_allowed("PUT"):
            self.set_custom_default_headers()
            await self.handle_through_thing(Operations.writeproperty)
        self.finish()

    async def delete(self) -> None:
        if await self.is_method_allowed("DELETE"):
            self.set_custom_default_headers()
            await self.handle_through_thing(Operations.deleteproperty)
        self.finish()


class ActionHandler(RPCHandler):
    """handles action requests"""

    async def get(self) -> None:
        if await self.is_method_allowed("GET"):
            self.set_custom_default_headers()
            if self.message_id is not None:
                await self.handle_no_block_response()
            else:
                await self.handle_through_thing(Operations.invokeaction)
        self.finish()

    async def post(self) -> None:
        if await self.is_method_allowed("POST"):
            self.set_custom_default_headers()
            await self.handle_through_thing(Operations.invokeaction)
        self.finish()

    async def put(self) -> None:
        if await self.is_method_allowed("PUT"):
            self.set_custom_default_headers()
            await self.handle_through_thing(Operations.invokeaction)
        self.finish()

    async def delete(self) -> None:
        if await self.is_method_allowed("DELETE"):
            self.set_custom_default_headers()
            await self.handle_through_thing(Operations.invokeaction)
        self.finish()


class RWMultiplePropertiesHandler(ActionHandler):
    """handles read-write of multiple properties via an action"""

    def initialize(
        self,
        resource: ActionAffordance,
        config: Any,
        logger: structlog.stdlib.BoundLogger,
        metadata: Any = None,
        **kwargs,
    ) -> None:
        self.read_properties_resource = kwargs.get("read_properties_resource", None)
        self.write_properties_resource = kwargs.get("write_properties_resource", None)
        return super().initialize(resource, config, logger, metadata)

    async def get(self) -> None:
        if await self.is_method_allowed("GET"):
            self.set_custom_default_headers()
            self.resource = self.read_properties_resource
            if self.message_id is not None:
                await self.handle_no_block_response()
            else:
                await self.handle_through_thing(Operations.invokeaction)
        self.finish()

    async def post(self) -> None:
        if await self.is_method_allowed("POST"):
            self.set_status(405, "method not allowed, PUT instead")
        self.finish()

    async def put(self) -> None:
        if await self.is_method_allowed("PUT"):
            self.set_custom_default_headers()
            self.resource = self.write_properties_resource
            await self.handle_through_thing(Operations.invokeaction)
        self.finish()

    async def patch(self) -> None:
        if await self.is_method_allowed("PATCH"):
            self.set_custom_default_headers()
            self.resource = self.write_properties_resource
            await self.handle_through_thing(Operations.invokeaction)
        self.finish()


class EventHandler(BaseHandler):
    """handles events emitted by `Thing` and tunnels them as HTTP SSE"""

    def initialize(
        self,
        resource: InteractionAffordance | EventAffordance,
        config: Any,
        logger: structlog.stdlib.BoundLogger,
        metadata: Any = None,
    ) -> None:
        super().initialize(resource, config, logger, metadata)
        self.data_header = b"data: %s\n\n"

    def set_custom_default_headers(self) -> None:
        """
        sets default headers for event handling. The general headers are listed as follows:

        ```yaml
        Content-Type: text/event-stream
        Cache-Control: no-cache
        Connection: keep-alive
        Access-Control-Allow-Credentials: true # Possibly for cookie auth
        Access-Control-Allow-Origin: <client> # if CORS is enabled
        ```
        """
        self.set_header("Content-Type", "text/event-stream")
        self.set_header("Cache-Control", "no-cache")
        self.set_header("Connection", "keep-alive")
        super().set_custom_default_headers()

    async def get(self):
        """events are support only with GET method"""
        if await self.has_access_control():
            self.set_custom_default_headers()
            await self.handle_datastream()
        self.finish()

    async def options(self):
        """options for the resource"""
        if await self.has_access_control():
            self.set_status(204)
            self.set_custom_default_headers()
            self.set_access_control_allow_headers()
            self.set_header("Access-Control-Allow-Methods", "GET")
        self.finish()

    def receive_blocking_event(self, event_consumer: EventConsumer):
        """deprecated, but can make a blocking call in an async loop"""
        return event_consumer.receive(timeout=10000, deserialize=False)

    async def handle_datastream(self) -> None:
        """called by GET method and handles the event publishing"""
        try:
            event_consumer = self.thing.subscribe_event(self.resource)
            self.set_status(200)
        except Exception as ex:
            self.logger.error(f"error while subscribing to event - {str(ex)}")
            self.set_status(500, f"could not subscribe to event source from thing - {str(ex)}")
            self.write(Serializers.json.dumps({"exception": format_exception_as_json(ex)}))
            return

        while True:
            try:
                event_message = await event_consumer.receive(timeout=10000)
                if event_message:
                    payload = self.thing.get_response_payload(event_message)
                    self.write(self.data_header % payload.value)
                    self.logger.debug(f"new data scheduled to flush - {self.resource.name}")
                else:
                    self.logger.debug(f"found no new data - {self.resource.name}")
                await self.flush()  # flushes and handles heartbeat - raises StreamClosedError if client disconnects
            except StreamClosedError:
                break
            except Exception as ex:
                self.logger.error(f"error while pushing event - {str(ex)}")
                self.write(self.data_header % Serializers.json.dumps({"exception": format_exception_as_json(ex)}))


class JPEGImageEventHandler(EventHandler):
    """handles events with images with JPEG image data header"""

    def initialize(
        self,
        resource: InteractionAffordance | EventAffordance,
        config: Any,
        logger: structlog.stdlib.BoundLogger,
        metadata: Any = None,
    ) -> None:
        super().initialize(resource, config, logger, metadata)
        self.data_header = b"data:image/jpeg;base64,%s\n\n"


class PNGImageEventHandler(EventHandler):
    """handles events with images with PNG image data header"""

    def initialize(
        self,
        resource: InteractionAffordance | EventAffordance,
        config: Any,
        logger: structlog.stdlib.BoundLogger,
        metadata: Any = None,
    ) -> None:
        super().initialize(resource, config, logger, metadata)
        self.data_header = b"data:image/png;base64,%s\n\n"


class StopHandler(BaseHandler):
    """Stops the tornado HTTP server"""

    def initialize(
        self,
        config: Any,
        logger: structlog.stdlib.BoundLogger,
        owner_inst: Any,
    ) -> None:
        from . import HTTPServer  # noqa: F401
        from .config import RuntimeConfig  # noqa: F401

        self.config = config  # type: RuntimeConfig
        self.logger = logger.bind(path=self.request.path)
        self.allowed_clients = self.config.allowed_clients
        self.security_schemes = self.config.security_schemes
        self.server = owner_inst  # type: HTTPServer

    async def post(self):
        if not await self.has_access_control():
            return
        try:
            self.set_custom_default_headers()
            # Stop the Tornado server
            origin = self.request.headers.get("Origin")
            self.logger.info(f"stopping HTTP server as per client request from {origin}, scheduling a stop message...")
            # create a task in current loop
            eventloop = get_current_async_loop()
            eventloop.create_task(self.server.async_stop())
            # dont call it in sequence, its not clear whether its designed for that
            self.set_status(204, "ok")
        except Exception as ex:
            self.logger.error(f"error while stopping HTTP server - {str(ex)}")
            self.set_status(500, f"error while stopping HTTP server - {str(ex)}")
        self.finish()


class LivenessProbeHandler(BaseHandler):
    """Liveness probe handler"""

    def initialize(
        self,
        config: Any,
        logger: structlog.stdlib.BoundLogger,
        owner_inst: Any = None,
    ) -> None:
        from . import HTTPServer  # noqa: F401
        from .config import RuntimeConfig  # noqa: F401

        self.config = config  # type: RuntimeConfig
        self.logger = logger.bind(path=self.request.path)
        self.server = owner_inst  # type: HTTPServer

    async def get(self):
        self.set_status(200, "ok")
        self.set_custom_default_headers()
        self.finish()


class ReadinessProbeHandler(BaseHandler):
    """Readiness probe handler"""

    def initialize(
        self,
        config: Any,
        logger: structlog.stdlib.BoundLogger,
        owner_inst: Any = None,
    ) -> None:
        from . import HTTPServer  # noqa: F401
        from .config import RuntimeConfig  # noqa: F401

        self.config = config  # type: RuntimeConfig
        self.logger = logger.bind(path=self.request.path)
        self.server = owner_inst  # type: HTTPServer

    async def get(self):
        self.set_custom_default_headers()
        try:
            if len(self.server._disconnected_things) > 0:
                raise RuntimeError("some things are disconnected, retry later")
            replies = await self.server.zmq_client_pool.async_execute_in_all_things(
                objekt="ping",
                operation="invokeaction",
            )
            if not all(reply.body[0].deserialize() is None for thing_id, reply in replies.items()):
                self.set_status(500, "not all things are ready")
            else:
                self.set_status(200, "ok")
                # self.write({id: "ready" for id in replies.keys()})
        except Exception as ex:
            self.logger.error(f"error while checking readiness - {str(ex)}")
            self.set_status(500, f"error while checking readiness - {str(ex)}")
        self.finish()


class ThingDescriptionHandler(BaseHandler):
    """Thing Description handler"""

    def initialize(
        self,
        resource: InteractionAffordance | PropertyAffordance,
        config: Any,
        logger: structlog.stdlib.BoundLogger,
        owner_inst: Any = None,
        metadata: Any = None,
    ) -> None:
        super().initialize(
            resource=resource,
            config=config,
            logger=logger,
            metadata=metadata,
        )
        self.thing_description = self.config.thing_description_service(
            resource=resource,
            config=config,
            logger=logger,
            server=owner_inst,
        )

    async def get(self):
        if not await self.has_access_control():
            self.finish()
            return

        try:
            self.set_custom_default_headers()
            _, _, _, body = self.get_execution_parameters()
            body = body.deserialize() or dict()
            if not isinstance(body, dict):
                raise ValueError("request body must be or convertable to JSON when supplied as path parameters")

            ignore_errors = body.get("ignore_errors", False)
            skip_names = body.get("skip_names", [])
            authority = body.get("authority", None)
            use_localhost = body.get("use_localhost", False)

            TD = await self.thing_description.generate(
                ignore_errors=ignore_errors,
                skip_names=skip_names,
                use_localhost=use_localhost,
                authority=authority,
            )

            self.set_status(200, "ok")
            self.set_header("Content-Type", "application/json")
            self.write(TD)
        except Exception as ex:
            self.set_status(500, str(ex).replace("\n", " "))
        self.finish()



from typing import Any, Optional

import structlog
import zmq.asyncio

from pydantic import BaseModel, ConfigDict, model_validator

from ..config import global_config
from ..core import Thing
from ..core.zmq.brokers import (
    AsyncEventConsumer,
    AsyncZMQClient,
    MessageMappedZMQClientPool,
    PreserializedData,
    PreserializedEmptyByte,
    ResponseMessage,
    SerializableData,
    SerializableNone,
    ServerExecutionContext,
    ThingExecutionContext,
    default_server_execution_context,
    default_thing_execution_context,
)
from ..core.zmq.message import EMPTY_BYTE
from ..td.interaction_affordance import EventAffordance, PropertyAffordance
from ..utils import uuid_hex


class BrokerThing(BaseModel):
    """Repository Layer of a Thing over the internal message broker"""

    id: str
    """Thing ID"""
    server_id: str
    """ZMQ Server ID"""
    access_point: str
    """ZMQ Access Point"""

    TD: dict[str, Any] | None = None
    """ZMQ Thing Description"""

    req_rep_client: MessageMappedZMQClientPool | None = None
    """req-rep queue client"""
    event_client: AsyncEventConsumer | None = None
    """pub-sub queue client"""
    req_rep_socket_address: str = ""
    """req-rep socket address"""
    pub_sub_socket_address: str = ""
    """pub-sub socket address"""
    logger: structlog.stdlib.BoundLogger | None = None
    """Logger instance"""

    @model_validator(mode="before")
    def validate_access_point(cls, values):
        """Validates the access point format before setting."""
        access_point = values.get("access_point")
        if access_point is not None and access_point.upper() not in ["TCP", "IPC", "INPROC"]:
            raise ValueError("Access point must be 'TCP', 'IPC', or 'INPROC'")
        return values

    async def execute(
        self,
        objekt: str,
        operation: str,
        payload: SerializableData = SerializableNone,
        preserialized_payload: PreserializedData = PreserializedEmptyByte,
        server_execution_context: ServerExecutionContext = default_server_execution_context,
        thing_execution_context: ThingExecutionContext = default_thing_execution_context,
    ) -> ResponseMessage:
        """
        Executes a request-response operation on a `Thing`.

        Parameters
        ----------
        objekt: str
            The target object of the operation - name of property or action.
        operation: str
            The operation to perform (e.g., "readproperty", "writeproperty", "invokeaction")
        payload: SerializableData, optional
            The payload to send with the request
        preserialized_payload: PreserializedData, optional
            The preserialized payload to send with the request
        server_execution_context: ServerExecutionContext, optional
            The server execution context
        thing_execution_context: ThingExecutionContext, optional
            The thing execution context
        """
        if self.req_rep_client is None:
            raise RuntimeError("Not connected to broker")
        return await self.req_rep_client.async_execute(
            thing_id=self.id,
            objekt=objekt,
            operation=operation,
            payload=payload,
            preserialized_payload=preserialized_payload,
            server_execution_context=server_execution_context,
            thing_execution_context=thing_execution_context,
        )

    async def schedule(
        self,
        objekt: str,
        operation: str,
        payload: SerializableData = SerializableNone,
        preserialized_payload: PreserializedData = PreserializedEmptyByte,
        server_execution_context: ServerExecutionContext = default_server_execution_context,
        thing_execution_context: ThingExecutionContext = default_thing_execution_context,
    ) -> str:
        """
        Schedules a request-response operation on a `Thing` and returns a message ID.

        Parameters
        ----------
        objekt: str
            The target object of the operation - name of property or action.
        operation: str
            The operation to perform (e.g., "readproperty", "writeproperty", "invokeaction")
        payload: SerializableData, optional
            The payload to send with the request
        preserialized_payload: PreserializedData, optional
            The preserialized payload to send with the request
        server_execution_context: ServerExecutionContext, optional
            The server execution context
        thing_execution_context: ThingExecutionContext, optional
            The thing execution context

        Returns
        -------
        str
            The message ID of the scheduled request
        """
        if self.req_rep_client is None:
            raise RuntimeError("Not connected to broker")
        return await self.req_rep_client.async_send_request(
            thing_id=self.id,
            objekt=objekt,
            operation=operation,
            payload=payload,
            preserialized_payload=preserialized_payload,
            server_execution_context=server_execution_context,
            thing_execution_context=thing_execution_context,
        )

    async def oneway(
        self,
        objekt: str,
        operation: str,
        payload: SerializableData = SerializableNone,
        preserialized_payload: PreserializedData = PreserializedEmptyByte,
        server_execution_context: ServerExecutionContext = default_server_execution_context,
        thing_execution_context: ThingExecutionContext = default_thing_execution_context,
    ) -> None:
        """
        Sends a oneway request(no-response) operation on a `Thing`.

        Parameters
        ----------
        objekt: str
            The target object of the operation - name of property or action.
        operation: str
            The operation to perform (e.g., "readproperty", "writeproperty", "invokeaction")
        payload: SerializableData, optional
            The payload to send with the request
        preserialized_payload: PreserializedData, optional
            The preserialized payload to send with the request
        server_execution_context: ServerExecutionContext, optional
            The server execution context
        thing_execution_context: ThingExecutionContext, optional
            The thing execution context
        """
        if self.req_rep_client is None:
            raise RuntimeError("Not connected to broker")
        await self.req_rep_client.async_send_request(
            thing_id=self.id,
            objekt=objekt,
            operation=operation,
            payload=payload,
            preserialized_payload=preserialized_payload,
            server_execution_context=server_execution_context,
            thing_execution_context=thing_execution_context,
        )

    async def recv_response(
        self,
        message_id: str,
        timeout: int = 10000,
    ) -> ResponseMessage:
        """
        Receives a response for a previously scheduled request.

        Parameters
        ----------
        message_id: str
            The message ID of the scheduled request
        timeout: int, optional
            The timeout in milliseconds to wait for the response (default is 10000)

        Returns
        -------
        ResponseMessage
            The response message received
        """
        if self.req_rep_client is None:
            raise RuntimeError("Not connected to broker")
        return await self.req_rep_client.async_recv_response(
            thing_id=self.id,
            message_id=message_id,
            timeout=timeout,
        )

    def subscribe_event(self, resource: EventAffordance | PropertyAffordance) -> AsyncEventConsumer:
        """
        Subscribe to events from a `Thing` through the internal pub-sub broker.

        Parameters
        ----------
        resource: EventAffordance | PropertyAffordance
            The event or observable property to subscribe to

        Returns
        -------
        AsyncEventConsumer
            The event consumer for the subscribed events
        """
        event_consumer = AsyncEventConsumer(
            id=f"{resource.name}|EventTunnel|{uuid_hex()}",
            event_unique_identifier=f"{resource.thing_id}/{resource.name}",
            access_point=self.pub_sub_socket_address,
            context=global_config.zmq_context(),
        )
        event_consumer.subscribe()
        return event_consumer

    def set_req_rep_client(self, client: AsyncZMQClient | MessageMappedZMQClientPool) -> None:
        """Sets the req-rep client for this broker thing."""
        self.req_rep_client = client
        self.req_rep_socket_address = client.socket_address

    def set_event_consumer(self, client: AsyncEventConsumer) -> None:
        """Sets the pub-sub client for this broker thing."""
        self.event_client = client
        self.pub_sub_socket_address = client.socket_address

    def get_response_payload(self, zmq_response: ResponseMessage) -> PreserializedData | SerializableData:
        """
        Retrieves the payload from the ZMQ response message, does not necessarily deserialize it.
        Use this method to extract the payload from a response message.
        Multipart responses are not supported yet for protocol controllers (except ZMQ), so only one payload is returned.

        Parameters
        ----------
        zmq_response: ResponseMessage
            The ZMQ response message to extract the payload from

        Returns
        -------
        PreserializedData | SerializableData
            The extracted payload, either preserialized or serialized
        """
        # print("zmq_response - ", zmq_response)
        if zmq_response is None:
            raise RuntimeError("No last response available. Did you make an operation?")
        if zmq_response.preserialized_payload.value != EMPTY_BYTE:
            if zmq_response.payload.value != b"null":
                # our None return value comes like this, sufficient to check against that
                self.logger.warning(
                    "Multiple content types in response payload (multipart payloads) are currently not supported,"
                    + " only the preserialized payload will be written to the wire",
                    content_type=zmq_response.payload.content_type,
                    binary_value=zmq_response.payload.value,
                )
            # multiple content types are not supported yet, so we return only one payload
            return zmq_response.preserialized_payload
            # return payload, preserialized_payload
        return zmq_response.payload  # dont deseriablize, there is no need, just pass it on to the client

    model_config = ConfigDict(arbitrary_types_allowed=True)


thing_repository = dict()  # type: dict[str, BrokerThing]


async def consume_broker_queue(
    id: str,
    server_id: str,
    thing_id: str,
    access_point: str,
    context: Optional[zmq.asyncio.Context] = None,
    logger: Optional[structlog.stdlib.BoundLogger] = None,
    poll_timeout: int = 1000,
) -> tuple[AsyncZMQClient, dict[str, Any]]:
    """
    Connect to a running Thing via ZMQ INPROC and fetch its Thing Description.

    Parameters
    ----------
    id: str
        Unique identifier for the client.
    server_id: str
        The server ID to connect to.
    thing_id: str
        The Thing ID whose Thing Description (TD) is to be fetched.
    access_point: str
        The access point (e.g., "TCP", "WS", or a specific address).
    context: Optional[zmq.asyncio.Context], optional
        ZMQ context to use for the connection. If None, uses the global context.
    logger: Optional[structlog.stdlib.BoundLogger], optional
        Logger instance for logging events. If None, no logging is performed.
    poll_timeout: int, optional
        Poll timeout in milliseconds (default is 1000).

    Returns
    -------
    tuple[AsyncZMQClient, dict[str, Any]]
        A tuple containing the connected AsyncZMQClient and the fetched Thing Description as a dictionary.
    """
    from ..client.zmq.consumed_interactions import ZMQAction

    # create client
    client = AsyncZMQClient(
        id=id,
        server_id=server_id,
        access_point=access_point,
        context=context or global_config.zmq_context(),
        handshake=False,
        logger=logger,
        poll_timeout=poll_timeout,
    )
    # connect client
    client.handshake(10000)
    await client.handshake_complete(10000)

    # fetch ZMQ INPROC TD
    Thing.get_thing_model  # type: Action
    FetchTMAffordance = Thing.get_thing_model.to_affordance()
    FetchTMAffordance.override_defaults(thing_id=thing_id, name="get_thing_description")
    fetch_td = ZMQAction(
        resource=FetchTMAffordance,
        sync_client=None,
        async_client=client,
        logger=logger,
        owner_inst=None,
    )
    if isinstance(access_point, str) and len(access_point) in [3, 6]:
        access_point = access_point.upper()
    elif access_point.lower().startswith("tcp://"):
        access_point = "TCP"
    TD = await fetch_td.async_call(ignore_errors=True, protocol=access_point)  # type: dict[str, Any]
    return client, TD


def consume_broker_pubsub(id: str, access_point: str) -> AsyncEventConsumer:
    """
    Consume all events from the broker's pub-sub queue.

    Parameters
    ----------
    id: str
        Unique identifier for the event consumer
    access_point: str
        The qualified ZMQ address (`tcp://`, `ipc://`, `inproc://`) of the pub-sub broker
    """
    return AsyncEventConsumer(
        id=id or f"EventTunnel|{uuid_hex()}",
        event_unique_identifier="",
        access_point=access_point,
        context=global_config.zmq_context(),
    )



from typing import Any, Dict, List, Tuple

import pytest

from pydantic import BaseModel, ValidationError

from hololinked.utils import (
    get_input_model_from_signature,
    issubklass,
    pydantic_validate_args_kwargs,
)


def func_without_args():
    return 1


def func_with_annotations(a: int, b: int) -> int:
    return a + b


def func_with_missing_annotations(a: int, b):
    return a + b


def func_with_no_annotations(a, b):
    return a + b


def func_with_kwargs(a: int, b: int, **kwargs):
    return a + b


def func_with_annotated_kwargs(a: int, b: int, **kwargs: dict[str, int]):
    return a + b


def func_with_args(*args):
    return sum(args)


def func_with_annotated_args(*args: list[int]):
    return sum(args)


def func_with_args_and_kwargs(*args, **kwargs):
    return sum(args) + sum(kwargs.values())


def func_with_annotated_args_and_kwargs(*args: list[int], **kwargs: dict[str, int]):
    return sum(args) + sum(kwargs.values())


def test_func_without_args_model_none():
    model = get_input_model_from_signature(func_without_args)
    assert model is None


def test_01_model_func_with_annotations():
    model = get_input_model_from_signature(func_with_annotations)
    assert issubklass(model, BaseModel)
    assert model.model_fields["a"].annotation is int
    assert model.model_fields["b"].annotation is int
    assert len(model.model_fields) == 2
    assert model.model_config["extra"] == "forbid"


@pytest.mark.parametrize(
    "args,kwargs,raises,exmsg",
    [
        (None, {"a": 1, "b": 2}, None, None),
        (None, {"a": 1, "b": "2"}, ValidationError, None),
        (None, {"a": 1}, ValidationError, None),
        (None, {"a": 1, "b": 2, "c": 3}, ValueError, "Unexpected keyword arguments"),
        ((1, 2), None, None, None),
        ((1, "2"), None, ValidationError, None),
        ((1, 2, 3), None, ValueError, "Too many positional arguments"),
        ((1,), None, ValidationError, None),
        ((1,), {"b": 2}, None, None),
        ((1,), {"a": 2}, ValueError, "Multiple values for argument"),
        ((1, 2), {"c": 3}, ValueError, "Unexpected keyword arguments"),
        (("1", 2), {"c": 3}, ValueError, "Unexpected keyword arguments"),
    ],
)
def test_01_validation_func_with_annotations(args, kwargs, raises, exmsg):
    model = get_input_model_from_signature(func_with_annotations)
    if raises:
        with pytest.raises(raises) as ex:
            pydantic_validate_args_kwargs(model, args=args if args else (), kwargs=kwargs if kwargs else {})
        if exmsg:
            assert str(ex.value).startswith(exmsg)
    else:
        pydantic_validate_args_kwargs(model, args=args if args else (), kwargs=kwargs if kwargs else {})


def test_02_model_func_with_missing_annotations():
    model = get_input_model_from_signature(func_with_missing_annotations)
    assert issubklass(model, BaseModel)
    assert model.model_fields["a"].annotation is int
    assert model.model_fields["b"].annotation is Any
    assert len(model.model_fields) == 2
    assert model.model_config["extra"] == "forbid"


@pytest.mark.parametrize(
    "args,kwargs,raises,exmsg",
    [
        (None, {"a": 1, "b": 2}, None, None),
        (None, {"a": 1, "b": "2"}, None, None),
        (None, {"a": 2, "b": list()}, None, None),
        (None, {"a": "1", "b": "2"}, ValidationError, None),
        (None, {"a": list(), "b": dict()}, ValidationError, None),
        (None, {"a": 1}, ValidationError, None),
        (None, {"a": 1, "b": 2, "c": 3}, ValueError, "Unexpected keyword arguments"),
        ((1, 2), None, None, None),
        ((1, "2"), None, None, None),
        ((2, list()), None, None, None),
        (("1", "2"), None, ValidationError, None),
        ((list(), dict()), None, ValidationError, None),
        ((1, 2, 3), None, ValueError, "Too many positional arguments"),
        ((1,), None, ValidationError, None),
        ((1,), {"b": 2}, None, None),
        ((1,), {"b": "2"}, None, None),
        ((2,), {"b": list()}, None, None),
        ((1,), {"a": 2}, ValueError, "Multiple values for argument"),
        ((1, 2), {"c": 3}, ValueError, "Unexpected keyword arguments"),
        (("1", 2), {"c": 3}, ValueError, "Unexpected keyword arguments"),
    ],
)
def test_02_validation_func_with_missing_annotations(args, kwargs, raises, exmsg):
    model = get_input_model_from_signature(func_with_missing_annotations)
    if raises:
        with pytest.raises(raises) as ex:
            pydantic_validate_args_kwargs(model, args=args if args else (), kwargs=kwargs if kwargs else {})
        if exmsg:
            assert str(ex.value).startswith(exmsg)
    else:
        pydantic_validate_args_kwargs(model, args=args if args else (), kwargs=kwargs if kwargs else {})


def test_03_model_func_with_no_annotations():
    model = get_input_model_from_signature(func_with_no_annotations, model_for_empty_annotations=True)
    assert issubklass(model, BaseModel)
    assert model.model_fields["a"].annotation is Any
    assert model.model_fields["b"].annotation is Any
    assert len(model.model_fields) == 2
    assert model.model_config["extra"] == "forbid"


@pytest.mark.parametrize(
    "args,kwargs,raises,exmsg",
    [
        (None, {"a": 1, "b": 2}, None, None),
        (None, {"a": 1.2, "b": "2"}, None, None),
        (None, {"a": dict(), "b": list()}, None, None),
        (None, {"a": list()}, ValidationError, None),
        (None, {"b": dict()}, ValidationError, None),
        (None, {"a": 1, "b": 2, "c": 3}, ValueError, "Unexpected keyword arguments"),
        ((1, 2), None, None, None),
        ((1, "2"), None, None, None),
        ((dict(), list()), None, None, None),
        ((1,), {"b": 2}, None, None),
        ((1, 2, 3), None, ValueError, "Too many positional arguments"),
        ((dict(), list(), 3), None, ValueError, "Too many positional arguments"),
        ((1,), None, ValidationError, None),
        ((dict(),), None, ValidationError, None),
        ((1,), {"b": 2}, None, None),
        ((1.1,), {"b": "2"}, None, None),
        ((dict(),), {"b": list()}, None, None),
        ((1,), {"a": 2}, ValueError, "Multiple values for argument"),
        ((1, 2), {"c": 3}, ValueError, "Unexpected keyword arguments"),
        (("1", 2), {"c": 3}, ValueError, "Unexpected keyword arguments"),
    ],
)
def test_03_validation_func_with_no_annotations(args, kwargs, raises, exmsg):
    model = get_input_model_from_signature(func_with_no_annotations, model_for_empty_annotations=True)
    if raises:
        with pytest.raises(raises) as ex:
            pydantic_validate_args_kwargs(model, args=args if args else (), kwargs=kwargs if kwargs else {})
        if exmsg:
            assert str(ex.value).startswith(exmsg)
    else:
        pydantic_validate_args_kwargs(model, args=args if args else (), kwargs=kwargs if kwargs else {})


def test_03_no_model_func_with_no_annotations():
    model = get_input_model_from_signature(func_with_no_annotations)
    assert model is None


def test_04_model_func_with_kwargs():
    model = get_input_model_from_signature(func_with_kwargs)
    assert issubklass(model, BaseModel)
    assert model.model_fields["a"].annotation is int
    assert model.model_fields["b"].annotation is int
    assert len(model.model_fields) == 3
    assert model.model_config["extra"] == "forbid"


@pytest.mark.parametrize(
    "args,kwargs,raises,exmsg",
    [
        (None, {"a": 1, "b": 2}, None, None),
        (None, {"a": 1, "b": 2, "c": 3}, None, None),
        ((1, 2), {"c": "3"}, None, None),
        (None, {"a": 1, "b": "2"}, ValidationError, None),
        (None, {"a": 1, "b": "2", "c": "3"}, ValidationError, None),
        (None, {"a": 1}, ValidationError, None),
        (None, {"a": 1, "b": 2, "c": 3, "d": 4}, None, None),
        ((1, 2), None, None, None),
        ((1, "2"), None, ValidationError, None),
        (("1", 2), None, ValidationError, None),
        ((1, 2, 3), None, ValidationError, None),
        ((1,), None, ValidationError, None),
        ((1,), {"b": 2}, None, None),
        ((1,), {"b": 2, "c": 3}, None, None),
        ((1,), {"a": 2}, ValueError, "Multiple values for argument"),
        ((1, 2), {"a": 3}, ValueError, "Multiple values for argument"),
        ((1, 2), {"b": 3}, ValueError, "Multiple values for argument"),
    ],
)
def test_04_validation_func_with_kwargs(args, kwargs, raises, exmsg):
    model = get_input_model_from_signature(func_with_kwargs)
    if raises:
        with pytest.raises(raises) as ex:
            pydantic_validate_args_kwargs(model, args=args if args else (), kwargs=kwargs if kwargs else {})
        if exmsg:
            assert str(ex.value).startswith(exmsg)
    else:
        pydantic_validate_args_kwargs(model, args=args if args else (), kwargs=kwargs if kwargs else {})


def test_05_model_func_with_annotated_kwargs():
    model = get_input_model_from_signature(func_with_annotated_kwargs)
    assert issubklass(model, BaseModel)
    assert model.model_fields["a"].annotation is int
    assert model.model_fields["b"].annotation is int
    assert model.model_fields["kwargs"].annotation == dict[str, int]
    assert len(model.model_fields) == 3
    assert model.model_config["extra"] == "forbid"


@pytest.mark.parametrize(
    "args,kwargs,raises,exmsg",
    [
        (None, {"a": 1, "b": 2}, None, None),
        (None, {"a": 1, "b": 2, "c": 3}, None, None),
        ((1, 2), {"c": 3}, None, None),
        (None, {"a": 1, "b": "2"}, ValidationError, None),
        (None, {"a": 1, "b": 2, "c": "3"}, ValidationError, None),
        (None, {"a": 1, "b": 2, "c": list()}, ValidationError, None),
        (None, {"a": 1}, ValidationError, None),
        ((1, 2), None, None, None),
        ((1, "2"), None, ValidationError, None),
        ((dict(), 2), None, ValidationError, None),
        ((1, 2, 3), None, ValidationError, None),
        ((1,), None, ValidationError, None),
        ((1,), {"b": 2}, None, None),
        ((1,), {"b": 2, "c": 3}, None, None),
        ((1,), {"a": 2}, ValueError, "Multiple values for argument"),
        ((1, 2), {"a": 3}, ValueError, "Multiple values for argument"),
        ((1, 2), {"b": 3}, ValueError, "Multiple values for argument"),
        ((1, 2), {"a": list(), "c": 3}, ValueError, "Multiple values for argument"),
    ],
)
def test_05_validation_func_with_annotated_kwargs(args, kwargs, raises, exmsg):
    model = get_input_model_from_signature(func_with_annotated_kwargs)
    if raises:
        with pytest.raises(raises) as ex:
            pydantic_validate_args_kwargs(model, args=args if args else (), kwargs=kwargs if kwargs else {})
        if exmsg:
            assert str(ex.value).startswith(exmsg)
    else:
        pydantic_validate_args_kwargs(model, args=args if args else (), kwargs=kwargs if kwargs else {})


def test_06_model_func_with_args():
    model = get_input_model_from_signature(func_with_args, model_for_empty_annotations=True)
    assert issubklass(model, BaseModel)
    # assert model.model_fields["args"].annotation == tuple or model.model_fields["args"].annotation == Tuple
    assert len(model.model_fields) == 1
    assert model.model_config["extra"] == "forbid"


@pytest.mark.parametrize(
    "args,kwargs,raises,exmsg",
    [
        ((1, 2), None, None, None),
        (None, None, None, None),
        ((dict(),), None, None, None),
        (None, {"a": 1}, ValueError, "Unexpected keyword arguments"),
        ((1, 2), None, None, None),
        ((1,), {"a": 2}, ValueError, "Unexpected keyword arguments"),
        ((1, 2), {"c": 3}, ValueError, "Unexpected keyword arguments"),
    ],
)
def test_06_validation_func_with_args(args, kwargs, raises, exmsg):
    model = get_input_model_from_signature(func_with_args, model_for_empty_annotations=True)
    if raises:
        with pytest.raises(raises) as ex:
            pydantic_validate_args_kwargs(model, args=args if args else (), kwargs=kwargs if kwargs else {})
        if exmsg:
            assert str(ex.value).startswith(exmsg)
    else:
        pydantic_validate_args_kwargs(model, args=args if args else (), kwargs=kwargs if kwargs else {})


def test_06_no_model_func_with_args():
    model = get_input_model_from_signature(func_with_args)
    assert model is None


def test_07_model_func_with_annotated_args():
    model = get_input_model_from_signature(func_with_annotated_args)
    assert issubklass(model, BaseModel)
    # assert model.model_fields["args"].annotation == list[int]
    assert len(model.model_fields) == 1
    assert model.model_config["extra"] == "forbid"


@pytest.mark.parametrize(
    "args,kwargs,raises,exmsg",
    [
        (None, {"a": 1}, ValueError, "Unexpected keyword arguments"),
        (None, None, None, None),
        ((1, 2), None, None, None),
        ((1, "2"), None, ValidationError, None),
        ((dict(),), None, ValidationError, None),
    ],
)
def test_07_validation_func_with_annotated_args(args, kwargs, raises, exmsg):
    model = get_input_model_from_signature(func_with_annotated_args)
    if raises:
        with pytest.raises(raises) as ex:
            pydantic_validate_args_kwargs(model, args=args if args else (), kwargs=kwargs if kwargs else {})
        if exmsg:
            assert str(ex.value).startswith(exmsg)
    else:
        pydantic_validate_args_kwargs(model, args=args if args else (), kwargs=kwargs if kwargs else {})


def test_08_no_model_func_with_args_and_kwargs():
    model = get_input_model_from_signature(func_with_args_and_kwargs)
    assert model is None


def test_08_model_func_with_args_and_kwargs():
    model = get_input_model_from_signature(func_with_args_and_kwargs, model_for_empty_annotations=True)
    assert issubklass(model, BaseModel)
    assert model.model_fields["args"].annotation == Tuple or model.model_fields["args"].annotation is tuple
    assert model.model_fields["kwargs"].annotation == Dict[str, Any] or model.model_fields["kwargs"].annotation is dict
    assert len(model.model_fields) == 2
    assert model.model_config["extra"] == "forbid"


def test_08_model_func_with_annotated_args_and_kwargs_model():
    model = get_input_model_from_signature(func_with_annotated_args_and_kwargs)
    assert issubklass(model, BaseModel)
    assert model.model_fields["args"].annotation == List[int] or model.model_fields["args"].annotation == list[int]
    assert (
        model.model_fields["kwargs"].annotation == Dict[str, int]
        or model.model_fields["kwargs"].annotation == dict[str, int]
    )
    assert len(model.model_fields) == 2
    assert model.model_config["extra"] == "forbid"


if __name__ == "__main__":
    pytest.main([__file__, "-v", "-s"])



import jsonschema

from pydantic import BaseModel

from ..constants import JSON
from ..utils import json_schema_merge_args_to_kwargs, pydantic_validate_args_kwargs


class BaseSchemaValidator:  # type definition
    """
    Base class for all schema validators.
    Serves as a type definition.
    """

    def __init__(self, schema: JSON | BaseModel) -> None:
        self.schema = schema

    def validate(self, data) -> None:
        """validate the data against the schema"""
        raise NotImplementedError("validate method must be implemented by subclass")

    def validate_method_call(self, args, kwargs) -> None:
        """validate the method call against the schema"""
        raise NotImplementedError("validate_method_call method must be implemented by subclass")


class JSONSchemaValidator(BaseSchemaValidator):
    """
    JSON schema validator according to standard python JSON schema.
    Somewhat slow, consider `FastJSONSchemaValidator` (`pip install fastjsonschema`) or
    pydantic annotation based validation if possible.
    """

    def __init__(self, schema) -> None:
        """
        Parameters
        ----------
        schema: JSON
            The JSON schema to validate against
        """
        jsonschema.Draft7Validator.check_schema(schema)
        super().__init__(schema)
        self.validator = jsonschema.Draft7Validator(schema)

    def validate(self, data) -> None:
        self.validator.validate(data)

    def validate_method_call(self, args, kwargs) -> None:
        if len(args) > 0:
            kwargs = json_schema_merge_args_to_kwargs(self.schema, args, kwargs)
        self.validate(kwargs)

    def json(self) -> JSON:
        """allows JSON (de-)serializable of the instance itself"""
        return self.schema

    def __get_state__(self):
        return self.schema

    def __set_state__(self, schema):
        return JSONSchemaValidator(schema)


class PydanticSchemaValidator(BaseSchemaValidator):
    """Schema validator according to pydantic models"""

    def __init__(self, schema: BaseModel) -> None:
        """
        Parameters
        ----------
        schema: BaseModel
            The pydantic model to validate against
        """
        super().__init__(schema)
        self.validator = schema.model_validate

    def validate(self, data) -> None:
        self.validator(data)

    def validate_method_call(self, args, kwargs) -> None:
        pydantic_validate_args_kwargs(self.schema, args, kwargs)

    def json(self) -> JSON:
        """allows JSON (de-)serializable of the instance itself"""
        return self.schema.model_dump_json()

    def __get_state__(self):
        return self.json()

    def __set_state__(self, schema: JSON):
        return PydanticSchemaValidator(BaseModel(**schema))


try:
    import fastjsonschema

    class FastJSONSchemaValidator(BaseSchemaValidator):
        """JSON schema validator according to fast JSON schema"""

        # Useful for performance with dictionary based schema specification
        # which msgspec has no built in support. Normally, for speed,
        # one should try to use msgspec's struct concept.

        def __init__(self, schema: JSON) -> None:
            super().__init__(schema)
            self.validator = fastjsonschema.compile(schema)

        def validate(self, data) -> None:
            """validates and raises exception when failed directly to the caller"""
            self.validator(data)

        def validate_method_call(self, args, kwargs) -> None:
            if len(args) > 0:
                kwargs = json_schema_merge_args_to_kwargs(self.schema, args, kwargs)
            self.validate(kwargs)

        def json(self) -> JSON:
            """allows JSON (de-)serializable of the instance itself"""
            return self.schema

        def __get_state__(self):
            return self.schema

        def __set_state__(self, schema):
            return FastJSONSchemaValidator(schema)

except ImportError:
    pass



import time

from typing import Any, Generator

import pytest

from hololinked.client.abstractions import SSE
from hololinked.client.factory import ClientFactory
from hololinked.client.proxy import ObjectProxy
from hololinked.utils import uuid_hex


try:
    from .things import TestThing
    from .utils import fake
except ImportError:
    from things import TestThing
    from utils import fake


@pytest.fixture(scope="class")
def access_point(request) -> str:
    return "INPROC"


@pytest.fixture(scope="class")
def thing(access_point) -> Generator[TestThing, None, None]:
    thing_id = f"test-thing-{uuid_hex()}"
    thing = TestThing(id=thing_id)
    thing.run_with_zmq_server(forked=True, access_points=[access_point])
    yield thing
    thing.rpc_server.stop()


@pytest.fixture(scope="class")
def thing_model(thing: TestThing) -> dict[str, Any]:
    return thing.get_thing_model(ignore_errors=True).json()


@pytest.fixture(scope="class")
def client(thing: TestThing, access_point: str) -> Generator[ObjectProxy, None, None]:
    client = ClientFactory.zmq(
        server_id=thing.id,
        thing_id=thing.id,
        access_point=access_point.replace("*", "localhost"),
        ignore_TD_errors=True,
    )
    yield client
    # client.close()


@pytest.mark.asyncio(loop_scope="class")
class TestRPC_E2E:
    """End-to-end tests for RPC"""

    def test_01_creation_and_handshake(self, client: ObjectProxy, thing_model: dict[str, Any]):
        assert isinstance(client, ObjectProxy)
        assert len(client.properties) + len(client.actions) + len(client.events) >= (
            len(thing_model["properties"]) + len(thing_model["actions"]) + len(thing_model["events"])
        )

    @pytest.mark.parametrize(
        "payload",
        [
            pytest.param(fake.text(max_nb_chars=100), id="text"),
            pytest.param(fake.sentence(), id="sentence"),
            pytest.param(fake.json(), id="json"),
        ],
    )
    def test_02_invoke_action_manual(self, client: ObjectProxy, payload: Any):
        assert client.invoke_action("action_echo", payload) == payload

    @pytest.mark.parametrize(
        "payload",
        [
            pytest.param(fake.chrome(), id="chrome"),
            pytest.param(fake.sha256(), id="sha256"),
            pytest.param(fake.address(), id="address"),
        ],
    )
    def test_03_invoke_action_dot_notation(self, client: ObjectProxy, payload: Any):
        assert client.action_echo(payload) == payload

    @pytest.mark.parametrize(
        "payload",
        [
            pytest.param(fake.random_number(), id="random-number"),
            pytest.param(fake.random_int(), id="random-int"),
        ],
    )
    def test_04_invoke_action_oneway(self, client: ObjectProxy, payload: Any):
        assert client.invoke_action("set_non_remote_number_prop", payload, oneway=True) is None
        assert client.get_non_remote_number_prop() == payload

    @pytest.mark.parametrize(
        "payload",
        [
            pytest.param(
                fake.pylist(20, value_types=[int, float, str, bool]),
                id="pylist-explicit-types",
            ),
        ],
    )
    def test_05_invoke_action_noblock(self, client: ObjectProxy, payload: Any):
        noblock_msg_id = client.invoke_action("action_echo", payload, noblock=True)
        assert isinstance(noblock_msg_id, str)
        assert client.invoke_action("action_echo", fake.pylist(20, value_types=[int, float, str, bool])) == fake.last
        assert client.invoke_action("action_echo", fake.pylist(10, value_types=[int, float, str, bool])) == fake.last
        assert client.read_reply(noblock_msg_id) == payload

    def test_06_read_property_manual(self, client: ObjectProxy):
        assert isinstance(client.read_property("number_prop"), (int, float))
        assert isinstance(client.read_property("string_prop"), str)
        assert client.read_property("selector_prop") in TestThing.selector_prop.objects

    @pytest.mark.parametrize(
        "prop, payload",
        [
            pytest.param("number_prop", fake.random_number(), id="random-number"),
            pytest.param(
                "selector_prop",
                TestThing.selector_prop.objects[fake.random_int(0, len(TestThing.selector_prop.objects) - 1)],
                id="selector-value",
            ),
            pytest.param(
                "observable_list_prop",
                fake.pylist(25, value_types=[int, float, str, bool]),
                id="observable-list",
            ),
        ],
    )
    def test_07_write_property_manual(self, client: ObjectProxy, prop: str, payload: Any):
        client.write_property(prop, payload)
        assert client.read_property(prop) == payload

    def test_08_read_property_dot_notation(self, client: ObjectProxy):
        assert isinstance(client.number_prop, (int, float))
        assert isinstance(client.string_prop, str)
        assert client.selector_prop in TestThing.selector_prop.objects

    def test_09_write_property_dot_notation(self, client: ObjectProxy):
        client.number_prop = fake.random_number()
        assert client.number_prop == fake.last
        client.selector_prop = TestThing.selector_prop.objects[
            fake.random_int(0, len(TestThing.selector_prop.objects) - 1)
        ]
        assert client.selector_prop == TestThing.selector_prop.objects[fake.last]
        client.observable_list_prop = fake.pylist(25, value_types=[int, float, str, bool])
        assert client.observable_list_prop == fake.last

    @pytest.mark.parametrize(
        "prop, payload",
        [
            pytest.param("number_prop", fake.random_number(), id="random-number"),
            pytest.param(
                "selector_prop",
                TestThing.selector_prop.objects[fake.random_int(0, len(TestThing.selector_prop.objects) - 1)],
                id="selector-value",
            ),
            pytest.param(
                "observable_list_prop",
                fake.pylist(25, value_types=[int, float, str, bool]),
                id="observable-list",
            ),
        ],
    )
    def test_10_write_property_oneway(self, client: ObjectProxy, prop: str, payload: Any):
        client.write_property(prop, payload, oneway=True)
        assert client.read_property(prop) == payload

    def test_11_read_property_noblock(self, client: ObjectProxy):
        noblock_msg_id = client.read_property("number_prop", noblock=True)
        assert isinstance(noblock_msg_id, str)
        assert client.read_property("selector_prop") in TestThing.selector_prop.objects
        assert isinstance(client.read_property("string_prop"), str)
        assert client.read_reply(noblock_msg_id) == client.number_prop

    def test_12_write_property_noblock(self, client: ObjectProxy):
        noblock_msg_id = client.write_property("number_prop", fake.random_number(), noblock=True)
        assert isinstance(noblock_msg_id, str)
        assert client.read_property("number_prop") == fake.last
        assert client.read_reply(noblock_msg_id) is None

    def test_13_error_handling(self, client: ObjectProxy):
        client.string_prop = "world"
        assert client.string_prop == "world"
        with pytest.raises(ValueError):
            client.string_prop = "WORLD"
        with pytest.raises(TypeError):
            client.int_prop = "5"
        with pytest.raises(AttributeError):
            _ = client.non_remote_number_prop

    def test_14_rw_multiple_properties(self, client: ObjectProxy):
        client.write_multiple_properties(number_prop=15, string_prop="foobar")
        assert client.number_prop == 15
        assert client.string_prop == "foobar"
        client.int_prop = 5
        client.selector_prop = "b"
        client.number_prop = -15
        props = client.read_multiple_properties(names=["selector_prop", "int_prop", "number_prop", "string_prop"])
        assert props["selector_prop"] == "b"
        assert props["int_prop"] == 5
        assert props["number_prop"] == -15
        assert props["string_prop"] == "foobar"

    def test_15_subscribe_event(self, client: ObjectProxy):
        results = []

        def cb(value: SSE):
            results.append(value)

        client.subscribe_event("test_event", cb)
        time.sleep(3)

        for i in range(10):
            client.push_events(total_number_of_events=1)
            time.sleep(1)
            if len(results) > 0:
                results.clear()
                break
        else:
            pytest.skip("No events received from server, probably due to OS level issues")

        client.push_events()
        time.sleep(3)
        assert len(results) > 0, "No events received"
        assert abs(len(results) - 100) < 3, f"Expected 100 events, got {len(results)}"
        client.unsubscribe_event("test_event")

    @pytest.mark.parametrize(
        "prop, prospective_values, op",
        [
            pytest.param(
                "observable_list_prop",
                [
                    [1, 2, 3, 4, 5],
                    ["a", "b", "c", "d", "e"],
                    [1, "a", 2, "b", 3],
                ],
                "write",
                id="observable-list-prop",
            ),
            pytest.param(
                "observable_readonly_prop",
                [1, 2, 3, 4, 5],
                "read",
                id="observable-readonly-prop",
            ),
        ],
    )
    def test_16_observe_properties(self, client: ObjectProxy, prop: str, prospective_values: Any, op: str):
        assert hasattr(client, f"{prop}_change_event")
        result = []
        attempt = 0

        def cb(value: SSE):
            nonlocal attempt
            result.append(value)
            attempt += 1

        client.observe_property(prop, cb)
        time.sleep(3)
        for value in prospective_values:
            if op == "read":
                _ = client.read_property(prop)
            else:
                client.write_property(prop, value)
        for _ in range(20):
            if attempt == len(prospective_values):
                break
            time.sleep(0.1)
        client.unobserve_property(prop)
        for index, res in enumerate(result):
            assert res.data == prospective_values[index]

    @pytest.mark.parametrize(
        "payload",
        [
            pytest.param(fake.text(max_nb_chars=100), id="text"),
            pytest.param(fake.sentence(), id="sentence"),
            pytest.param(fake.json(), id="json"),
        ],
    )
    async def test_17_async_invoke_action(self, client, payload):
        result = await client.async_invoke_action("action_echo", payload)
        assert result == payload

    async def test_18_async_read_property(self, client):
        assert isinstance(await client.async_read_property("number_prop"), (int, float))
        assert isinstance(await client.async_read_property("string_prop"), str)
        assert await client.async_read_property("selector_prop") in TestThing.selector_prop.objects

    @pytest.mark.parametrize(
        "prop, payload",
        [
            pytest.param("number_prop", fake.random_number(), id="random-number"),
            pytest.param(
                "selector_prop",
                TestThing.selector_prop.objects[fake.random_int(0, len(TestThing.selector_prop.objects) - 1)],
                id="selector-value",
            ),
            pytest.param(
                "observable_list_prop",
                fake.pylist(25, value_types=[int, float, str, bool]),
                id="observable-list",
            ),
        ],
    )
    async def test_19_async_write_property(self, client, prop, payload):
        await client.async_write_property(prop, payload)
        assert await client.async_read_property(prop) == payload


if __name__ == "__main__":
    pytest.main([__file__, "-v", "-s"])



from .controllers import (  # noqa: F401
    ActionHandler,
    BaseHandler,
    EventHandler,
    JPEGImageEventHandler,
    LivenessProbeHandler,
    PNGImageEventHandler,
    PropertyHandler,
    ReadinessProbeHandler,
    RPCHandler,
    RWMultiplePropertiesHandler,
    StopHandler,
    ThingDescriptionHandler,
)
from .server import HTTPServer  # noqa: F401



import socket
import ssl
import warnings

from copy import deepcopy
from typing import Any, Iterable, Type

import structlog

from tornado import ioloop
from tornado.httpserver import HTTPServer as TornadoHTTP1Server
from tornado.web import Application

from ...config import global_config
from ...constants import HTTP_METHODS
from ...core.actions import Action
from ...core.events import Event
from ...core.property import Property
from ...core.thing import Thing, ThingMeta
from ...core.zmq.brokers import MessageMappedZMQClientPool

# from tornado_http2.server import Server as TornadoHTTP2Server
from ...param.parameters import ClassSelector, IPAddress
from ...td import ActionAffordance, EventAffordance, PropertyAffordance
from ...utils import (
    get_current_async_loop,
    issubklass,
    pep8_to_dashed_name,
    run_callable_somehow,
)
from ..repository import thing_repository
from ..security import Security
from ..server import BaseProtocolServer
from .config import HandlerMetadata, RuntimeConfig
from .controllers import (
    ActionHandler,
    BaseHandler,
    EventHandler,
    LivenessProbeHandler,
    PropertyHandler,
    ReadinessProbeHandler,
    RWMultiplePropertiesHandler,
    StopHandler,
    ThingDescriptionHandler,
)
from .services import ThingDescriptionService


class HTTPServer(BaseProtocolServer):
    """
    HTTP(s) server to expose `Thing` over HTTP protocol. Supports HTTP 1.1.
    Use `add_thing`, or `add_property` or `add_action` or `add_event` methods to add things to the server.
    """

    address = IPAddress(default="0.0.0.0", doc="IP address")  # type: str
    # SAST(id='hololinked.server.http.HTTPServer.address', description='B104:hardcoded_bind_all_interfaces', tool='bandit')
    """IP address, especially to bind to all interfaces or not"""

    ssl_context = ClassSelector(
        class_=ssl.SSLContext,
        default=None,
        allow_None=True,
    )  # type: ssl.SSLContext | None
    """SSL context to provide encrypted communication"""

    config = ClassSelector(
        class_=RuntimeConfig,
        default=None,
        allow_None=True,
    )  # type: RuntimeConfig
    """Runtime configuration for the HTTP server. See `hololinked.server.http.config.RuntimeConfig` for details"""

    def __init__(
        self,
        *,
        port: int = 8080,
        address: str = "0.0.0.0",  # SAST(id='hololinked.server.http.HTTPServer.__init__.address', description='B104:hardcoded_bind_all_interfaces', tool='bandit')
        things: list[Thing] | None = None,
        # host: Optional[str] = None,
        logger: structlog.stdlib.BoundLogger | None = None,
        ssl_context: ssl.SSLContext | None = None,
        security_schemes: list[Security] | None = None,
        # protocol_version : int = 1, network_interface : str = 'Ethernet',
        allowed_clients: str | Iterable[str] | None = None,
        config: dict[str, Any] | None = None,
        **kwargs,
    ) -> None:
        """
        Parameters
        ----------
        port: int, default 8080
            the port at which the server should be run
        address: str, default 0.0.0.0
            IP address, use 0.0.0.0 to bind to all interfaces to expose the server to other devices in the network
            and 127.0.0.1 to bind only to localhost
        logger: structlog.stdlib.BoundLogger, optional
            structlog.stdlib.BoundLogger instance
        ssl_context: ssl.SSLContext
            SSL context to provide encrypted communication
        security_schemes: list[Security], optional
            list of security schemes to be used by the server. If None, no security scheme is used.
        allowed_clients: List[str]
            serves request and sets CORS only from these clients, other clients are reject with 403. Unlike pure CORS
            feature, the server resource is not even executed if the client is not an allowed client.
        **kwargs:
            additional keyword arguments for server configuration. Usually:

            - `property_handler`: `BaseHandler` | `PropertyHandler`, optional.
                custom web request handler for property read-write
            - `action_handler`: `BaseHandler` | `ActionHandler`, optional.
                custom web request handler for action
            - `event_handler`: `EventHandler` | `BaseHandler`, optional.
                custom event handler for sending HTTP SSE

            or RuntimeConfig attributes can be passed as keyword arguments.
        """
        default_config = dict(
            cors=global_config.ALLOW_CORS,
            property_handler=kwargs.get("property_handler", PropertyHandler),
            action_handler=kwargs.get("action_handler", ActionHandler),
            event_handler=kwargs.get("event_handler", EventHandler),
            thing_description_handler=kwargs.get("thing_description_handler", ThingDescriptionHandler),
            RW_multiple_properties_handler=kwargs.get("RW_multiple_properties_handler", RWMultiplePropertiesHandler),
            liveness_probe_handler=kwargs.get("liveness_handler", LivenessProbeHandler),
            readiness_probe_handler=kwargs.get("readiness_handler", ReadinessProbeHandler),
            stop_handler=kwargs.get("stop_handler", StopHandler),
            thing_description_service=kwargs.get("thing_description_service", ThingDescriptionService),
            thing_repository=kwargs.get("thing_repository", thing_repository),
            allowed_clients=allowed_clients,
            security_schemes=security_schemes,
        )
        default_config.update(config or dict())
        config = RuntimeConfig(**default_config)
        # need to be extended when more options are added
        super().__init__(
            port=port,
            address=address,
            logger=logger,
            ssl_context=ssl_context,
            config=config,
        )

        self._IP = f"{self.address}:{self.port}"
        if self.logger is None:
            self.logger = structlog.get_logger().bind(component="http-server", host=f"{self.address}:{self.port}")

        self.tornado_instance = None
        self.app = Application(
            handlers=[
                (
                    r"/stop",
                    self.config.stop_handler,
                    dict(config=self.config, logger=self.logger, owner_inst=self),
                ),
                (
                    r"/liveness",
                    self.config.liveness_probe_handler,
                    dict(config=self.config, logger=self.logger, owner_inst=self),
                ),
                (
                    r"/readiness",
                    self.config.readiness_probe_handler,
                    dict(config=self.config, logger=self.logger, owner_inst=self),
                ),
            ]
        )
        self.router = ApplicationRouter(self.app, self)

        self.zmq_client_pool = MessageMappedZMQClientPool(
            id=self._IP,
            server_ids=[],
            client_ids=[],
            handshake=False,
            poll_timeout=100,
        )
        self.add_things(*(things or []))

    def setup(self) -> None:
        """check if all the requirements are met before starting the server, auto invoked by listen()"""
        # Add only those code here that needs to be redone always before restarting the server.
        # One time creation attributes/activities must be in init

        # Comments are above the relevant lines, not below
        # 1. clear the event loop in case any pending tasks exist, also restarting with same
        # event loop is buggy, so we remove it.
        ioloop.IOLoop.clear_current()
        # 2. sets async loop for a non-possessing thread as well
        event_loop = get_current_async_loop()
        # 3. schedule the ZMQ client pool polling
        event_loop.create_task(self.zmq_client_pool.poll_responses())
        # self.zmq_client_pool.handshake(), NOTE - handshake better done upfront as we already poll_responses here
        # which will prevent handshake function to succeed (although handshake will be done)
        # 4. Expose via broker
        for thing in self.things:
            if not thing.rpc_server:
                raise ValueError(f"You need to expose thing {thing.id} via a RPCServer before trying to serve it")
            event_loop.create_task(
                self._instantiate_broker(
                    thing.rpc_server.id,
                    thing.id,
                    "INPROC",
                )
            )
        # 5. finally also get a reference of the same event loop from tornado
        self.tornado_event_loop = ioloop.IOLoop.current()

        self.tornado_instance = TornadoHTTP1Server(self.app, ssl_options=self.ssl_context)  # type: TornadoHTTP1Server

    async def start(self) -> None:
        self.setup()
        self.tornado_instance.listen(port=self.port, address=self.address)
        self.logger.info(f"started HTTP webserver at {self._IP}, ready to receive requests.")

    def stop(self, attempt_async_stop: bool = True) -> None:
        """
        Stop the HTTP server - unreliable, use `async_stop()` if possible.
        A stop handler at the path `/stop` with POST method is already implemented that invokes this
        method for the clients.

        Parameters
        ----------
        attempt_async_stop: bool, default `True`
            if `True`, attempts to run the `async_stop` method to close all connections gracefully.
        """
        if attempt_async_stop:
            run_callable_somehow(self.async_stop())
            return
        self.zmq_client_pool.stop_polling()
        if not self.tornado_instance:
            return
        self.tornado_instance.stop()
        run_callable_somehow(self.tornado_instance.close_all_connections())

    async def async_stop(self) -> None:
        """
        Stop the HTTP server. A stop handler at the path `/stop` with POST method is already implemented
        that invokes this method for the clients.
        """
        self.zmq_client_pool.stop_polling()
        if not self.tornado_instance:
            return
        try:
            self.tornado_instance.stop()
            await self.tornado_instance.close_all_connections()
        except Exception as ex:
            self.logger.error(
                "error while stopping tornado server, use stop() method "
                + f"from hololinked.server and do not reuse the port - {ex}"
            )

    def add_property(
        self,
        URL_path: str,
        property: Property | PropertyAffordance,
        http_methods: str | tuple[str, str, str] = ("GET", "PUT"),
        handler: BaseHandler | PropertyHandler = PropertyHandler,
        **kwargs,
    ) -> None:
        """
        Add a property to be accessible by HTTP.

        Parameters
        ----------
        URL_path: str
            URL path to access the property
        property: Property | PropertyAffordance
            Property (object) to be served or its JSON representation
        http_methods: Tuple[str, str, str]
            tuple of http methods to be used for read, write and delete. Use None or omit HTTP method for
            unsupported operations. For example - for readonly property use ('GET', None, None) or ('GET',)
        handler: BaseHandler | PropertyHandler, optional
            custom handler for the property, otherwise the default handler will be used
        kwargs: dict
            additional keyword arguments to be passed to the handler's __init__
        """
        if not isinstance(property, (Property, PropertyAffordance)):
            raise TypeError(f"property should be of type Property, given type {type(property)}")
        if not issubklass(handler, BaseHandler):
            raise TypeError(f"handler should be subclass of BaseHandler, given type {type(handler)}")
        if isinstance(property, Property):
            property = property.to_affordance()
        read_http_method = write_http_method = delete_http_method = None
        http_methods = self.router.adapt_http_methods(http_methods)
        if len(http_methods) == 1:
            read_http_method = http_methods[0]
        elif len(http_methods) == 2:
            read_http_method, write_http_method = http_methods
        elif len(http_methods) == 3:
            read_http_method, write_http_method, delete_http_method = http_methods
        if read_http_method != "GET":
            raise ValueError("read method should be GET")
        if write_http_method and write_http_method not in ["POST", "PUT"]:
            raise ValueError("write method should be POST or PUT")
        if delete_http_method and delete_http_method != "DELETE":
            raise ValueError("delete method should be DELETE")
        kwargs["resource"] = property
        kwargs["logger"] = self.logger
        kwargs["config"] = self.config
        kwargs["metadata"] = HandlerMetadata(http_methods=http_methods)
        self.router.add_rule(affordance=property, URL_path=URL_path, handler=handler, kwargs=kwargs)

    def add_action(
        self,
        URL_path: str,
        action: Action | ActionAffordance,
        http_method: str | None = "POST",
        handler: BaseHandler | ActionHandler = ActionHandler,
        **kwargs,
    ) -> None:
        """
        Add an action to be accessible by HTTP

        Parameters
        ----------
        URL_path: str
            URL path to access the action
        action: Action | ActionAffordance
            Action (object) to be served or its JSON representation
        http_method: str
            http method to be used for the action
        handler: BaseHandler | ActionHandler, optional
            custom handler for the action
        kwargs: dict
            additional keyword arguments to be passed to the handler's __init__
        """
        if not isinstance(action, (Action, ActionAffordance)):
            raise TypeError(f"Given action should be of type Action or ActionAffordance, given type {type(action)}")
        if not issubklass(handler, BaseHandler):
            raise TypeError(f"handler should be subclass of BaseHandler, given type {type(handler)}")
        http_methods = self.router.adapt_http_methods(http_method)
        if isinstance(action, Action):
            action = action.to_affordance()  # type: ActionAffordance
        kwargs["resource"] = action
        kwargs["config"] = self.config
        kwargs["logger"] = self.logger
        kwargs["metadata"] = HandlerMetadata(http_methods=http_methods)
        self.router.add_rule(affordance=action, URL_path=URL_path, handler=handler, kwargs=kwargs)

    def add_event(
        self,
        URL_path: str,
        event: Event | EventAffordance | PropertyAffordance,
        handler: BaseHandler | EventHandler = EventHandler,
        **kwargs,
    ) -> None:
        """
        Add an event to be accessible by HTTP server; only GET method is supported for events.

        Parameters
        ----------
        URL_path: str
            URL path to access the event
        event: Event | EventAffordance
            Event (object) to be served or its JSON representation
        handler: BaseHandler | EventHandler, optional
            custom handler for the event
        kwargs: dict
            additional keyword arguments to be passed to the handler's __init__
        """
        if not isinstance(event, (Event, EventAffordance)) and (
            not isinstance(event, PropertyAffordance) or not event.observable
        ):
            raise TypeError(f"event should be of type Event or EventAffordance, given type {type(event)}")
        if not issubklass(handler, BaseHandler):
            raise TypeError(f"handler should be subclass of BaseHandler, given type {type(handler)}")
        if isinstance(event, Event):
            event = event.to_affordance()
        kwargs["resource"] = event
        kwargs["config"] = self.config
        kwargs["logger"] = self.logger
        kwargs["metadata"] = HandlerMetadata(http_methods=("GET",))
        self.router.add_rule(affordance=event, URL_path=URL_path, handler=handler, kwargs=kwargs)

    def add_thing(self, thing: Thing) -> None:
        self.router.add_thing(thing)
        self.things.append(thing)

    def __hash__(self):
        return hash(self._IP)

    def __eq__(self, other):
        if not isinstance(other, HTTPServer):
            return False
        return self._IP == other._IP

    def __str__(self):
        return f"{self.__class__.__name__}(address={self.address}, port={self.port})"


class ApplicationRouter:
    """
    Covering implementation (as in - a layer on top) of the application router to
    add rules to the tornado application. Not a real router, which is taken care of
    by the tornado application automatically.
    """

    def __init__(self, app: Application, server: HTTPServer) -> None:
        self.app = app
        self.server = server
        self.logger = server.logger.bind(component="http-router")
        self._pending_rules = []
        self._rules = dict()  # type: dict[str, Any]

    # can add a single property, action or event rule
    def add_rule(
        self,
        affordance: PropertyAffordance | ActionAffordance | EventAffordance,
        URL_path: str,
        handler: Type[BaseHandler],
        kwargs: dict[str, Any],
    ) -> None:
        """
        Add rule to the application router. Note that this method will replace existing rules and can duplicate
        endpoints for an affordance without checks (i.e. you could technically add two different endpoints for the
        same affordance).

        Parameters
        ----------
        affordance: PropertyAffordance | ActionAffordance | EventAffordance
            the interaction affordance for which the rule is being added
        URL_path: str
            URL path to access the affordance
        handler: Type[BaseHandler]
            handler class to be used for the affordance
        kwargs: dict[str, Any]
            additional keyword arguments to be passed to the handler's __init__
        """
        for rule in self.app.wildcard_router.rules:
            if rule.matcher == URL_path:
                warnings.warn(
                    f"URL path {URL_path} already exists in the router -"
                    + f" replacing it for {affordance.what} {affordance.name}",
                    category=UserWarning,
                )
        for rule in self._pending_rules:
            if rule[0] == URL_path:
                warnings.warn(
                    f"URL path {URL_path} already exists in the pending rules -"
                    + f" replacing it for {affordance.what} {affordance.name}",
                    category=UserWarning,
                )
        if getattr(affordance, "thing_id", None):
            if not URL_path.startswith(f"/{affordance.thing_id}"):
                warnings.warn(
                    f"URL path {URL_path} does not start with the thing id {affordance.thing_id},"
                    + f" adding it to the path, new path = {f'/{affordance.thing_id}{URL_path}'}. "
                    + " This warning can be usually safely ignored."
                )
                URL_path = f"/{affordance.thing_id}{URL_path}"
            self.logger.info(
                f"adding rule for {affordance.what} {affordance.name} at path {URL_path}"
                + f" for thing id {affordance.thing_id}"
            )
            self.app.wildcard_router.add_rules([(URL_path, handler, kwargs)])
        elif affordance.thing_cls is not None:
            self.logger.info(
                f"adding pending rule for {affordance.what} {affordance.name} at path {URL_path}"
                + f" for thing class {affordance.thing_cls.__name__}, probably no forms exist yet."
            )
            self._pending_rules.append((URL_path, handler, kwargs))
        else:
            raise RuntimeError("object has no thing id or thing class associated with it, cannot add rule")
        """
        for handler based tornado rule matcher, the Rule object has following
        signature
        
        def __init__(
            self,
            matcher: "Matcher",
            target: Any,
            target_kwargs: Optional[Dict[str, Any]] = None,
            name: Optional[str] = None,
        ) -> None:

        matcher - based on route
        target - handler
        target_kwargs - given to handler's initialize
        name - ...

        len == 2 tuple is route + handler
        len == 3 tuple is route + handler + target kwargs
    
        so we give (path, BaseHandler, {'resource' : PropertyAffordance, 'owner' : self})
        
        path is extracted from interaction affordance name or given by the user
        BaseHandler is the base handler of this package for interaction affordances
        resource goes into target kwargs which is needed for the handler to work correctly
        """

    # can add multiple properties, actions and events at once
    def add_interaction_affordances(
        self,
        properties: Iterable[PropertyAffordance],
        actions: Iterable[ActionAffordance],
        events: Iterable[EventAffordance],
        thing_id: str = None,
    ) -> None:
        """
        Can add multiple properties, actions and events at once to the application router.
        Calls `add_rule` method internally for each affordance.

        Parameters
        ----------
        properties: Iterable[PropertyAffordance]
            list of properties to be added
        actions: Iterable[ActionAffordance]
            list of actions to be added
        events: Iterable[EventAffordance]
            list of events to be added
        thing_id: str, optional
            thing id to be prefixed to the URL path of each property, action, and event.
            If the thing_id is not provided, then the rule will be in pending state and not exposed
            until a thing instance with the given thing_id is added to the server.
        """
        for property in properties:
            if property in self:
                continue
            route = self.adapt_route(property.name)
            if property.thing_id is not None:
                path = f"/{property.thing_id}{route}"
            self.server.add_property(
                URL_path=path,
                property=property,
                http_methods=("GET",) if property.readOnly else ("GET", "PUT"),
                # if prop.fdel is None else ('GET', 'PUT', 'DELETE')
                handler=self.server.config.property_handler,
            )
            if property.observable:
                self.server.add_event(
                    URL_path=f"{path}/change-event",
                    event=property,
                    handler=self.server.config.event_handler,
                )
        for action in actions:
            if action in self:
                continue
            elif action.name == "get_thing_model":
                continue
            route = self.adapt_route(action.name)
            if action.thing_id is not None:
                path = f"/{action.thing_id}{route}"
            self.server.add_action(URL_path=path, action=action, handler=self.server.config.action_handler)
        for event in events:
            if event in self:
                continue
            route = self.adapt_route(event.name)
            if event.thing_id is not None:
                path = f"/{event.thing_id}{route}"
            self.server.add_event(URL_path=path, event=event, handler=self.server.config.event_handler)

        # thing model handler
        get_thing_model_action = next((action for action in actions if action.name == "get_thing_model"), None)
        self.server.add_action(
            URL_path=f"/{thing_id}/resources/wot-tm" if thing_id else "/resources/wot-tm",
            action=get_thing_model_action,
            http_method=("GET",),
        )

        # thing description handler
        get_thing_description_action = deepcopy(get_thing_model_action)
        get_thing_description_action.override_defaults(name="get_thing_description")
        self.server.add_action(
            URL_path=f"/{thing_id}/resources/wot-td" if thing_id else "/resources/wot-td",
            action=get_thing_description_action,
            http_method=("GET",),
            handler=self.server.config.thing_description_handler,
            owner_inst=self.server,
        )

        # RW multiple properties handler
        read_properties = Thing._get_properties.to_affordance(Thing)
        write_properties = Thing._set_properties.to_affordance(Thing)
        read_properties.override_defaults(thing_id=get_thing_model_action.thing_id)
        write_properties.override_defaults(thing_id=get_thing_model_action.thing_id)
        self.server.add_action(
            URL_path=f"/{thing_id}/properties" if thing_id else "/properties",
            action=read_properties,
            http_method=("GET", "PUT", "PATCH"),
            handler=self.server.config.RW_multiple_properties_handler,
            read_properties_resource=read_properties,
            write_properties_resource=write_properties,
        )

    # can add an entire thing instance at once
    def add_thing(self, thing: Thing) -> None:
        """
        internal method to add a thing instance to be served by the HTTP server. Iterates through the
        interaction affordances and adds a route for each property, action and event.
        """
        # Prepare affordance lists with error handling (single loop)
        if not isinstance(thing, Thing):
            raise TypeError(f"thing should be of type Thing, unknown type given - {type(thing)}")
        TM = thing.get_thing_model(ignore_errors=True).json()
        properties, actions, events = [], [], []
        for prop in TM.get("properties", dict()).keys():
            affordance = PropertyAffordance.from_TD(prop, TM)
            affordance.override_defaults(thing_id=thing.id, thing_cls=thing.__class__, owner=thing)
            properties.append(affordance)
        for action in TM.get("actions", dict()).keys():
            affordance = ActionAffordance.from_TD(action, TM)
            affordance.override_defaults(thing_id=thing.id, thing_cls=thing.__class__, owner=thing)
            actions.append(affordance)
        for event in TM.get("events", dict()).keys():
            affordance = EventAffordance.from_TD(event, TM)
            affordance.override_defaults(thing_id=thing.id, thing_cls=thing.__class__, owner=thing)
            events.append(affordance)
        self._resolve_rules(thing.id, thing.__class__)
        self.add_interaction_affordances(
            properties,
            actions,
            events,
            thing_id=thing.id,
        )

    def _resolve_rules(
        self,
        thing_id: str,
        thing_cls: ThingMeta,
    ) -> None:
        """
        Process the pending rules and add them to the application router.
        Rules become pending only when a property, action or event has a thing class associated
        but no thing instance.
        """
        pending_rules = self._pending_rules
        self._pending_rules = []
        for rule in pending_rules:
            affordance = rule[2].get("resource", None)  # type: PropertyAffordance | ActionAffordance | EventAffordance
            if not affordance or affordance.owner != thing_cls:
                self._pending_rules.append(rule)
                continue
            affordance.override_defaults(thing_cls=thing_cls, thing_id=thing_id)
            URL_path, handler, kwargs = rule
            URL_path = f"/{thing_id}{URL_path}"
            rule = (URL_path, handler, kwargs)
            self.add_rule(affordance=affordance, URL_path=URL_path, handler=handler, kwargs=kwargs)

    def __contains__(
        self,
        item: str | Property | Action | Event | PropertyAffordance | ActionAffordance | EventAffordance,
    ) -> bool:
        """
        Check if the item is in the application router.
        Not exact for torando's rules when a string is provided for the URL path,
        as you need to provide the Matcher object
        """
        if isinstance(item, str):
            for rule in self.app.wildcard_router.rules:
                if rule.matcher == item:
                    return True
                if hasattr(rule.matcher, "regex") and rule.matcher.regex.match(item):
                    return True
            for rule in self._pending_rules:
                if rule[0] == item:
                    return True
        elif isinstance(item, (Property, Action, Event)):
            item = item.to_affordance()
        if isinstance(item, (PropertyAffordance, ActionAffordance, EventAffordance)):
            for rule in self.app.wildcard_router.rules:
                if rule.target_kwargs.get("resource", None) == item:
                    return True
            for rule in self._pending_rules:
                if rule[2].get("resource", None) == item:
                    return True
        return False

    def get_href_for_affordance(self, affordance, authority: str = None, use_localhost: bool = False) -> str:
        """
        Get the full URL path for the affordance in the application router.

        Parameters
        ----------
        affordance: PropertyAffordance | ActionAffordance | EventAffordance
            the interaction affordance for which the URL path is to be retrieved
        authority: str, optional
            authority (protocol + host + port) to be used in the URL path. If None, the machine's hostname is used.
        use_localhost: bool, default `False`
            if `True`, localhost is used in the basepath instead of the server's hostname.

        Returns
        -------
        str
            full URL path for the affordance
        """
        if affordance not in self:
            raise ValueError(f"affordance {affordance} not found in the application router")
        for rule in self.app.wildcard_router.rules:
            if rule.target_kwargs.get("resource", None) == affordance:
                path = str(rule.matcher.regex.pattern).rstrip("$")
                return f"{self.get_basepath(authority, use_localhost)}{path}"

    def get_injected_dependencies(self, affordance) -> dict[str, Any]:
        """Get the target kwargs for the affordance in the application router"""
        if affordance not in self:
            raise ValueError(f"affordance {affordance} not found in the application router")
        for rule in self.app.wildcard_router.rules:
            if rule.target_kwargs.get("resource", None) == affordance:
                return rule.target_kwargs
        for rule in self._pending_rules:
            if rule[2].get("resource", None) == affordance:
                return rule[2]
        raise ValueError(f"affordance {affordance} not found in the application router rules")

    def get_basepath(self, authority: str = None, use_localhost: bool = False) -> str:
        """
        Get the basepath of the server.

        Parameters
        ----------
        authority: str, optional
            authority (protocol + host + port) to be used in the basepath. If None, the machine's hostname is used.
        use_localhost: bool, default `False`
            if `True`, localhost is used in the basepath instead of the server's hostname.
        """
        if authority:
            return authority
        protocol = "https" if self.server.ssl_context else "http"
        port = f":{self.server.port}" if self.server.port != 80 else ""
        if not use_localhost:
            return f"{protocol}://{socket.gethostname()}{port}"
        if self.server.address == "0.0.0.0" or self.server.address == "127.0.0.1":
            # SAST(id='hololinked.server.http.ApplicationRouter.get_basepath', description='B104:hardcoded_bind_all_interfaces', tool='bandit')
            return f"{protocol}://127.0.0.1{port}"
        elif self.server.address == "::":
            return f"{protocol}://[::1]{port}"
        return f"{protocol}://localhost{port}"

    basepath = property(fget=get_basepath, doc="basepath of the server")

    def adapt_route(self, interaction_affordance_name: str) -> str:
        """adapt the URL path to default conventions"""
        if interaction_affordance_name == "get_thing_model":
            return "/resources/wot-tm"
        return f"/{pep8_to_dashed_name(interaction_affordance_name)}"

    def adapt_http_methods(self, http_methods: Any):
        """comply the supplied HTTP method to the router to a tuple and check if the method is supported"""
        if isinstance(http_methods, str):
            http_methods = (http_methods,)
        if not isinstance(http_methods, tuple):
            raise TypeError("http_method should be a tuple")
        for method in http_methods:
            if method not in HTTP_METHODS.__members__.values() and method is not None:
                raise ValueError(f"method {method} not supported")
        return http_methods

    def print_rules(self) -> None:
        """
        Print the rules in the application router.
        prettytable is used if available, otherwise a simple print is done.
        """
        try:
            from prettytable import PrettyTable

            table = PrettyTable()
            table.field_names = ["URL Path", "Handler", "Resource Name"]

            for rule in self.app.wildcard_router.rules:
                table.add_row(
                    [
                        rule.matcher,
                        rule.target.__name__,
                        getattr(rule.target_kwargs.get("resource"), "name", "N/A"),
                    ]
                )
            for rule in self._pending_rules:
                table.add_row([rule[0], rule[1].__name__, rule[2]["resource"].name])
            print(table)
        except ImportError:
            print("Application Router Rules:")
            for rule in self.app.wildcard_router.rules:
                print(rule)
            for rule in self._pending_rules:
                print(rule[0], rule[2]["resource"].name)


__all__ = [HTTPServer.__name__]



from typing import Any, Optional

from pydantic import ConfigDict, Field

from ..core import Thing
from ..core.state_machine import BoundFSM
from .base import Schema
from .data_schema import DataSchema
from .interaction_affordance import (
    ActionAffordance,
    EventAffordance,
    PropertyAffordance,
)
from .metadata import VersionInfo


class ThingModel(Schema):
    """
    Thing Model as per W3C WoT Thing Description v1.1

    [Specification](https://www.w3.org/TR/wot-thing-description11/) <br>
    [UML Diagram](https://docs.hololinked.dev/UML/PDF/ThingModel.pdf) <br>
    """

    context: list[str | dict[str, str]] = Field(["https://www.w3.org/2022/wot/td/v1.1"], alias="@context")
    type: Optional[str | list[str]] = None
    id: str = None
    title: str = None
    description: Optional[str] = None
    version: Optional[VersionInfo] = None
    created: Optional[str] = None
    modified: Optional[str] = None
    support: Optional[str] = None
    base: Optional[str] = None
    properties: dict[str, DataSchema] = Field(default_factory=dict)
    actions: dict[str, ActionAffordance] = Field(default_factory=dict)
    events: dict[str, EventAffordance] = Field(default_factory=dict)

    model_config = ConfigDict(extra="allow")

    def __init__(
        self,
        instance: "Thing",
        allow_loose_schema: Optional[bool] = False,
        ignore_errors: bool = False,
        skip_names: Optional[list[str]] = [],
    ) -> None:
        super().__init__()
        self.instance = instance
        self.allow_loose_schema = allow_loose_schema
        self.ignore_errors = ignore_errors
        self.skip_names = skip_names or []

    def generate(self) -> "ThingModel":
        """populate the thing model"""
        self.id = self.instance.id
        self.title = self.instance.__class__.__name__
        self.context = ["https://www.w3.org/2022/wot/td/v1.1"]
        # default value of context is not being picked up although we only use exclude_unset=True
        if self.instance.__doc__:
            self.description = Schema.format_doc(self.instance.__doc__)
        self.properties = dict()
        self.actions = dict()
        self.events = dict()
        self.add_interaction_affordances()
        return self

    def produce(self) -> Thing:
        """produce a Thing instance from the Thing Model, not implemented yet"""
        raise NotImplementedError("This will be implemented in a future release for an API first approach")

    # not the best code and logic, but works for now
    skip_properties: list[str] = ["expose", "thing_description", "GUI", "object_info"]
    """list of property names to skip when generating the TD"""

    skip_actions: list[str] = [
        Thing._add_property.name,
        Thing._get_properties.name,
        Thing._get_properties_in_db.name,
        Thing._set_properties.name,
        "get_postman_collection",
        "get_our_thing_model",
    ]
    """list of action names to skip when generating the TD"""

    skip_events: list[str] = []
    """list of event names to skip when generating the TD"""

    def add_interaction_affordances(self) -> None:
        """add interaction affordances to thing model"""
        for affordance, items, affordance_cls, skip_list in [
            ["properties", self.instance.properties.remote_objects.items(), PropertyAffordance, self.skip_properties],
            ["actions", self.instance.actions.descriptors.items(), ActionAffordance, self.skip_actions],
            ["events", self.instance.events.plain.items(), EventAffordance, self.skip_events],
        ]:
            for name, obj in items:
                if name in skip_list or name in self.skip_names:
                    continue
                if (
                    name == "state"
                    and affordance == "properties"
                    and (
                        not hasattr(self.instance, "state_machine")
                        or not isinstance(self.instance.state_machine, BoundFSM)
                    )
                ):
                    continue
                try:
                    affordance_dict = getattr(self, affordance)
                    affordance_dict[name] = affordance_cls.generate(obj, self.instance)
                except Exception as ex:
                    if not self.ignore_errors:
                        raise ex from None
                    self.instance.logger.error(f"Error while generating schema for {name} - {ex}")

    def model_dump(self, **kwargs) -> dict[str, Any]:
        """Return the JSON representation of the schema"""

        def dump_value(value):
            nonlocal kwargs
            if hasattr(value, "model_dump"):
                return value.model_dump(**kwargs)
            elif isinstance(value, dict):
                return {k: dump_value(v) for k, v in value.items()}
            elif isinstance(value, list):
                return [dump_value(v) for v in value]
            elif isinstance(value, tuple):
                return tuple(dump_value(v) for v in value)
            else:
                return value

        result = {}
        for field in self.__class__.model_fields:
            if field in self.skip_keys:
                continue
            if not hasattr(self, field) or getattr(self, field) is None:
                continue
            if field in [
                "instance",
                "skip_keys",
                "skip_properties",
                "skip_actions",
                "skip_events",
                "skip_names",
                "ignore_errors",
                "allow_loose_schema",
            ]:
                continue
            value = getattr(self, field)
            result[field] = dump_value(value)
        return result



class ReplyNotArrivedError(Exception):
    """Exception raised when a reply is not received in time."""

    pass


class BreakLoop(Exception):
    """raise and catch to exit a loop from within another function or method"""

    pass



import copy
import logging
import logging.handlers
import sys
import types

from typing import Any

import structlog

from structlog.dev import KeyValueColumnFormatter


default_label_formatter = None


def normalize_component_name(_, __, event_dict: dict[str, Any]) -> dict[str, Any]:
    """cast component name to upper case and format event message with it"""

    global default_label_formatter
    component = event_dict.pop("component", "")
    if default_label_formatter:
        component_label = f"{default_label_formatter('component', component.upper())} " if component else ""
    else:
        component_label = f"[{component.upper()}] " if component else ""
    event_dict["event"] = f"{component_label}{event_dict.get('event', '')}"
    return event_dict


def setup_logging(
    log_level: int = logging.INFO,
    colored_logs: bool = False,
    log_file: str = None,
    **kwargs,
) -> None:
    """
    Setup structured logging using structlog. Not a flexible setup, except the values configurable in `global_config`,
    Override the entire function if you want a different logging configuration by monkey patching this method.

    Parameters
    ----------
    log_level: int
        logging level to use
    colored_logs: bool
        whether to use colored logs in console, usually harder to pick it up in fluentd
    log_file: str
        optional log file to log into
    **kwargs
        additional keyword arguments

        - `rotate_log_files`: `bool`, whether to rotate log files daily (default True)
        - `logfile_backup_count`: `int`, number of backup log files to keep (default 14)
    """
    handlers = []
    if log_file:
        if kwargs.get("rotate_log_files", True):
            filehandler = logging.handlers.TimedRotatingFileHandler(
                log_file,
                when="midnight",
                backupCount=kwargs.get("logfile_backup_count", 14),
            )
        else:
            filehandler = logging.FileHandler(log_file)
        filehandler.setFormatter(logging.Formatter("%(message)s"))
        handlers.append(filehandler)
    iostream_handler = logging.StreamHandler(sys.stdout)
    iostream_handler.setFormatter(logging.Formatter("%(message)s"))
    handlers.append(iostream_handler)

    logging.basicConfig(level=log_level, handlers=handlers, force=True)

    global default_label_formatter
    console_renderer = structlog.dev.ConsoleRenderer(colors=colored_logs)
    for column in console_renderer.columns:
        if column.key == "logger_name" and isinstance(column.formatter, KeyValueColumnFormatter):
            default_label_formatter = copy.deepcopy(column.formatter)
            default_label_formatter.key_style = None

    structlog.configure(
        processors=[
            structlog.contextvars.merge_contextvars,
            structlog.processors.add_log_level,
            structlog.processors.StackInfoRenderer(),
            structlog.dev.set_exc_info,
            structlog.processors.TimeStamper(fmt="%d-%m-%YT%H:%M:%S.%fZ"),
            normalize_component_name,
            console_renderer,
        ],
        logger_factory=structlog.stdlib.LoggerFactory(),
        context_class=dict,
        cache_logger_on_first_use=True,
        wrapper_class=structlog.make_filtering_bound_logger(log_level),
    )

    import asyncio  # noqa: F401

    asyncio_log = structlog.get_logger().bind(component="library|asyncio")
    for name, module in sys.modules.items():
        if name.startswith("asyncio.") and isinstance(module, types.ModuleType):
            if hasattr(module, "logger"):
                module.logger = asyncio_log

    try:
        import httpx  # noqa: F401

        # httpx_log = structlog.get_logger().bind(component="library|httpx")
        logging.getLogger("httpcore").setLevel(logging.WARNING if log_level <= logging.WARNING else log_level)
        logging.getLogger("httpx").setLevel(logging.WARNING if log_level <= logging.WARNING else log_level)
    except ImportError:
        pass

    try:
        import tornado.log

        tornado_log = structlog.get_logger().bind(component="library|tornado")
        tornado.log.access_log = tornado_log
        tornado.log.app_log = tornado_log
        tornado.log.gen_log = tornado_log
    except ImportError:
        pass



from typing import Any, Generator

import pytest

from hololinked.client import ClientFactory, ObjectProxy
from hololinked.server import stop
from hololinked.utils import uuid_hex


try:
    from .test_11_rpc_e2e import TestRPC_E2E as BaseRPC_E2E  # noqa: F401
    from .test_11_rpc_e2e import client, thing, thing_model  # noqa: F401
    from .test_14_protocols_http import hostname_prefix, wait_until_server_ready
    from .things import TestThing
except ImportError:
    from test_11_rpc_e2e import TestRPC_E2E as BaseRPC_E2E  # noqa: F401
    from test_11_rpc_e2e import client, thing, thing_model  # noqa: F401
    from test_14_protocols_http import hostname_prefix, wait_until_server_ready
    from things import TestThing


@pytest.fixture(scope="class")
def port() -> int:
    return 63000


@pytest.fixture(scope="class")
def thing(port: int) -> Generator[TestThing, None, None]:
    thing = TestThing(id=f"test-thing-{uuid_hex()}", serial_number="simulation")
    print()  # TODO, can be removed when tornado logs respect level
    thing.run_with_http_server(
        address="127.0.0.1",
        port=port,
        forked=True,
        print_welcome_message=False,
        config=dict(cors=True),
    )
    wait_until_server_ready(port=port)
    yield thing
    stop()


@pytest.fixture(scope="class")
def thing_model(thing: TestThing) -> dict[str, Any]:
    return thing.get_thing_model(ignore_errors=True).json()


@pytest.fixture(scope="class")
def td_endpoint(thing: TestThing, port: int) -> str:
    return f"{hostname_prefix}:{port}/{thing.id}/resources/wot-td"


@pytest.fixture(scope="class")
def client(td_endpoint: str) -> "ObjectProxy":
    return ClientFactory.http(url=td_endpoint, ignore_TD_errors=True)


@pytest.mark.asyncio(loop_scope="class")
class TestHTTP_E2E(BaseRPC_E2E):
    def test_14_rw_multiple_properties(self, client: ObjectProxy):
        pass


if __name__ == "__main__":
    pytest.main([__file__, "-v", "-s"])



from typing import Literal

from pydantic import BaseModel, ConfigDict, Field, model_validator
from pydantic.types import SecretStr, StrictInt, StrictStr


class SQLDBConfig(BaseModel):
    """Configuration validator for SQL databases for PostgreSQL and MySQL"""

    provider: Literal["postgresql", "mysql"] = "postgresql"
    """Database provider, postgresql or mysql"""
    host: StrictStr = "localhost"
    """PostgreSQL server host"""
    port: StrictInt = 5432
    """server port, default 5432"""
    database: StrictStr = "hololinked"
    """database name, default hololinked"""
    user: StrictStr = "hololinked"
    """user name, default hololinked, recommended not to use admin user"""
    password: SecretStr = Field(default=SecretStr(""), repr=False)
    """user password, default empty"""
    dialect: Literal["asyncpg", "psycopg", "asyncmy", "mysqldb", ""] = ""
    """dialect to use, default psycopg for postgresql"""
    uri: SecretStr = ""
    """Full database URI, overrides other settings, default empty"""

    model_config = ConfigDict(extra="forbid")

    @property
    def URL(self) -> str:
        if self.uri:
            return self.uri.get_secret_value()
        if self.provider == "postgresql":
            return f"postgresql{'+' + self.dialect if self.dialect else ''}://{self.user}:{self.password.get_secret_value()}@{self.host}:{self.port}/{self.database}"
        return f"mysql{'+' + self.dialect if self.dialect else ''}://{self.user}:{self.password.get_secret_value()}@{self.host}:{self.port}/{self.database}"

    @model_validator(mode="after")
    def _at_least_one(self):
        if not self.uri and (not self.host or not self.port or not self.database or not self.user or not self.password):
            raise ValueError("Provide either database URI or all of 'host', 'port', 'database', 'user', and 'password'")
        return self


class SQLiteConfig(BaseModel):
    """Configuration validator for SQLite database"""

    provider: Literal["sqlite"] = "sqlite"
    """Database provider, only sqlite is supported"""
    dialect: SecretStr = "pysqlite"
    """dialect to use, aiosqlite for async, pysqlite for sync"""
    file: str = ""
    """SQLite database file, default is empty string, which leads to an DB with name of thing ID"""
    in_memory: bool = False
    """Use in-memory SQLite database, default False as it is not persistent"""
    uri: SecretStr = ""
    """Full database URI, overrides other settings, default empty"""

    model_config = ConfigDict(extra="forbid")

    @property
    def URL(self) -> str:
        if self.uri:
            return self.uri.get_secret_value()
        if self.in_memory:
            return f"sqlite+{self.dialect}:///:memory:"
        elif self.file:
            return f"sqlite+{self.dialect}:///{self.file}"
        raise NotImplementedError("Either 'uri' or 'file' or 'in_memory' must be provided for SQLiteConfig")


class MongoDBConfig(BaseModel):
    """Configuration validator for MongoDB database"""

    provider: Literal["mongo"] = "mongo"
    """Database provider, only mongo is supported"""
    host: StrictStr = "localhost"
    """MongoDB server host"""
    port: StrictInt = 27017
    """server port, default 27017"""
    database: StrictStr = "hololinked"
    """database name, default hololinked"""
    user: StrictStr = ""
    """user name, default empty, recommended not to use admin user"""
    password: SecretStr = ""
    """user password, default empty"""
    authSource: StrictStr = ""
    """authentication source database, default empty"""
    uri: SecretStr = ""
    """Full database URI, overrides other settings, default empty"""

    model_config = ConfigDict(extra="forbid")

    @property
    def URL(self) -> str:
        if self.uri:
            return self.uri.get_secret_value()
        if self.user and self.password:
            if self.authSource:
                return f"mongodb://{self.user}:{self.password.get_secret_value()}@{self.host}:{self.port}/?authSource={self.authSource}"
            return f"mongodb://{self.user}:{self.password.get_secret_value()}@{self.host}:{self.port}/"
        return f"mongodb://{self.host}:{self.port}/"

    @model_validator(mode="after")
    def _at_least_one(self):
        if not self.uri and (not self.host or not self.port or not self.database or not self.user or not self.password):
            raise ValueError("Provide either database URI or all of 'host', 'port', 'database', 'user', and 'password'")
        return self



import typing

from enum import IntEnum, StrEnum

import zmq


# types
JSONSerializable = typing.Union[str, int, float, bool, None, typing.Dict[str, typing.Any], typing.List]
JSON = typing.Dict[str, JSONSerializable]

byte_types = (bytes, bytearray, memoryview)

# decorator constants
# naming
USE_OBJECT_NAME: str = "USE_OBJECT_NAME"
# state machine
ANY_STATE: str = "ANY_STATE"
UNSPECIFIED: str = "UNSPECIFIED"
# types


class ResourceTypes(StrEnum):
    """Exposed resource types"""

    PROPERTY = "PROPERTY"
    ACTION = "ACTION"
    EVENT = "EVENT"
    THING = "THING"


class REGEX(StrEnum):
    """common regexes"""

    states = "[A-Za-z_]+[A-Za-z_ 0-9]*"
    url = r"[\-a-zA-Z0-9@:%._\/\+~#=]{1,256}"


class HTTP_METHODS(StrEnum):
    """currently supported HTTP request methods"""

    GET = "GET"
    POST = "POST"
    PUT = "PUT"
    DELETE = "DELETE"
    PATCH = "PATCH"
    # OPTIONS = 'OPTIONS'


class ZMQ_TRANSPORTS(StrEnum):
    """
    supported ZMQ transport protocols - TCP, IPC, INPROC

    TCP - needs socket address additional specified like tcp://0.0.0.0:{port}
    IPC - within python, multiprocess applications, can be autogenerated
    INPROC - within python, multithreaded applications, can be autogenerated
    """

    TCP = "TCP"
    IPC = "IPC"
    INPROC = "INPROC"


class ZMQSocketType(IntEnum):
    """General ZMQ socket types"""

    PAIR = zmq.PAIR
    PUB = zmq.PUB
    SUB = zmq.SUB
    REQ = zmq.REQ
    REP = zmq.REP
    DEALER = zmq.DEALER
    ROUTER = zmq.ROUTER
    PULL = zmq.PULL
    PUSH = zmq.PUSH
    XPUB = zmq.XPUB
    XSUB = zmq.XSUB
    STREAM = zmq.STREAM
    # Add more socket types as needed


ZMQ_EVENT_MAP = {}
"""Built-in ZMQ events"""

for name in dir(zmq):
    if name.startswith("EVENT_"):
        value = getattr(zmq, name)
        ZMQ_EVENT_MAP[value] = name


class Operations(StrEnum):
    """WoT operations (both standard and non-standard)"""

    readproperty = "readproperty"
    writeproperty = "writeproperty"
    deleteproperty = "deleteproperty"
    observeproperty = "observeproperty"

    readmultipleproperties = "readmultipleproperties"
    writemultipleproperties = "writemultipleproperties"
    readallproperties = "readallproperties"
    writeallproperties = "writeallproperties"

    invokeaction = "invokeaction"
    subscribeevent = "subscribeevent"
    unsubscribevent = "unsubscribevent"


__all__ = [HTTP_METHODS.__name__, ZMQ_TRANSPORTS.__name__, Operations.__name__]
